import time
import copy
from typing import Optional, Union

import numpy as np
import pandas as pd

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from scipy import optimize
from scipy.stats import loguniform

from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.model_selection import ParameterSampler
from sklearn import preprocessing

from pybalance.utils import (
    MatchingData,
    split_target_pool,
    BaseBalanceCalculator,
    BalanceCalculator,
)

import logging

logger = logging.getLogger(__name__)


def _check_fitted(matcher):
    if matcher.best_match is None:
        raise (ValueError, "Matcher has not been fitted!")


class PropensityScoreMatcher:
    """
    Use a propensity score model to match two populations. The Matcher searches
    randomly over hyperparameters for the propensity score model and selects the
    match that performs best according to the given optimization objective.

    :param matching_data: Data containing pool and target populations to be
        matched.

    :param objective: Matching objective to optimize in hyperparameter search.
        Can be a string referring to any balance calculator known to
        utils.balance_calculators.BalanceCalculator or an instance of
        BaseBalanceCalculator.

    :param caliper: If defined, restricts matches to those patients with
        propensity scores within the caliper of each other. Note that caliper
        matching may lead to a loss of patients in the target population if no
        patient in the pool exists within the specified caliper. Should be in
        (0, 1].

    :param max_iter: Maximum number of hyperparameters to try before returning
        the best match.

    :param time_limit: Restrict hyperparameter search based on time. No new
        model will be trained after time_limit seconds have passed since
        matching began.

    :param method: Method to use for propensity score matching. Can be either
        'greedy' or 'linear_sum_assignment'. The former method is locally
        optimal and globally sub-optimial; the latter globally optimial but far
        more compute intensive. For large problems, use greedy.

    :param verbose: Flag to indicate whether to print diagnositic information
        during training.
    """

    DEFAULT_HYPERPARAM_SPACE = {
        LogisticRegression: {
            "C": loguniform(a=0.01, b=100),
            "penalty": ["l1", "l2"],
            "solver": ["saga"],
            "fit_intercept": [True, False],
            "max_iter": [500],
        },
        SGDClassifier: {
            "penalty": ["l1", "l2", "elasticnet"],
            "alpha": loguniform(a=0.02, b=20),
            "fit_intercept": [True, False],
            "early_stopping": [True, False],
            "loss": ["log_loss", "modified_huber"],
            "max_iter": [1500],
            "class_weight": [None, "balanced"],
        },
    }

    POOL = 0
    TARGET = 1

    def __init__(
        self,
        matching_data: MatchingData,
        objective: Union[str, BaseBalanceCalculator] = "beta",
        caliper: Optional[float] = None,
        max_iter: int = 50,
        time_limit: float = 60 * 5,
        method: str = "greedy",
        verbose: bool = True,
    ):
        self.caliper = caliper
        self.max_iter = max_iter
        self.time_limit = time_limit
        self.method = method
        self.hyperparam_space = self.DEFAULT_HYPERPARAM_SPACE
        self.verbose = verbose

        self.matching_data = matching_data.copy()
        self.target, self.pool = split_target_pool(matching_data)
        if isinstance(objective, str):
            self.balance_calculator = BalanceCalculator(self.matching_data, objective)
            self.objective = objective
        else:
            self.balance_calculator = objective
            self.objective = self.balance_calculator.name
        self._train_preprocessors(self.matching_data)
        self._reset_best_match()

    def get_params(self):
        params = ["objective", "caliper", "max_iter", "time_limit", "method"]
        return dict((p, getattr(self, p)) for p in params)

    def _reset_best_match(self):
        self.best_match = None
        self.best_match_idx = None
        self.best_score = np.inf
        self.best_model = None
        self.best_ps = None
        self.solution_time = 0

    def _describe_best_match(self):
        _check_fitted(self)
        logger.info("Best propensity score match found:")
        logger.info(f'\tModel: {str(self.best_model).split("(")[0]}')
        for key, val in self.best_params.items():
            logger.info(f"\t* {key}: {val}")
        logger.info(f"\tScore ({self.balance_calculator.name}): {self.best_score:.4f}")
        logger.info(f"\tSolution time: {self.solution_time:.3f} min")

    def _update_best_match(
        self, clf, params, match, score, ps_pool, ps_target, solution_time
    ):
        self.best_score = score
        self.best_match = match
        self.best_model = clf
        self.best_params = params
        self.best_ps = {"pool": ps_pool, "target": ps_target}
        self.solution_time = solution_time

        if self.verbose:
            self._describe_best_match()

    def get_best_match(self):
        _check_fitted(self)
        if self.verbose:
            self._describe_best_match()

        return self.best_match

    def match(self) -> MatchingData:
        """
        Match populations passed during __init__(). Returns MatchingData
        instance containing the matched pool and target populations.
        """
        t0 = time.time()

        # Search over hyperparameter search space to find a propensity score
        # that optimizes the given objective
        X, y = self._preprocess_data_for_sklearn(self.matching_data)
        hyperparams = self._get_hyperparams(self.max_iter)
        for i, (model, params) in enumerate(hyperparams):
            if (time.time() - t0) > self.time_limit:
                logger.warning("Time limit exceeded. Stopping early.")
                break

            clf = model(**params)
            logger.info(
                f'Training model {str(clf).split("(")[0]} (iter {i + 1}/{self.max_iter}, {(time.time() - t0)/60:.3f} min) ...'
            )

            clf.fit(X, y)
            ps_pool, ps_target = self.get_propensity_score(clf, self.matching_data)

            pool_matches, target_matches = propensity_score_match(
                ps_pool, ps_target, method=self.method, caliper=self.caliper
            )
            pool = self.pool.iloc[pool_matches]
            target = self.target.iloc[target_matches]
            match = MatchingData(
                pd.concat([pool, target]),
                headers=self.matching_data.headers,
                population_col=self.matching_data.population_col,
            )

            score = self.balance_calculator.distance(pool)

            if score < self.best_score:
                solution_time = (time.time() - t0) / 60
                self._update_best_match(
                    clf, params, match, score, ps_pool, ps_target, solution_time
                )

        return self.get_best_match()

    def _train_preprocessors(self, matching_data):
        self.standard_scaler = preprocessing.StandardScaler()
        X = self.balance_calculator.preprocessor.transform(matching_data)
        self.standard_scaler.fit(X.data[X.headers["all"]])

    def _preprocess_data_for_sklearn(self, matching_data):
        """
        Preprocess matching data to prepare it for estimating a propensity score.
        """
        matching_data = self.balance_calculator.preprocessor.transform(matching_data)
        target, pool = split_target_pool(matching_data)

        # construct feature data in a single array
        X = pd.concat([pool, target])[matching_data.headers["all"]]
        X = self.standard_scaler.transform(X)

        # construct target variable
        y = np.array([self.POOL] * len(pool) + [self.TARGET] * len(target))

        return X, y

    def _get_hyperparams(self, n_iter):
        hyperparams = []
        n_models = len(self.hyperparam_space)
        for model, params in self.hyperparam_space.items():
            hyperparams.extend(
                [
                    (model, p)
                    for p in ParameterSampler(params, n_iter=int(n_iter / n_models))
                ]
            )

        np.random.shuffle(hyperparams)

        return hyperparams

    def get_propensity_score(
        self, clf: BaseEstimator = None, matching_data: MatchingData = None
    ) -> tuple:
        """
        Returns the estimated propensity scores for the pool and target
        populations in a MatchingData object as predicted by a given
        propensity score model. If no model is passed, the method uses the
        best_model computed during matching. If no data are passed, the
        method uses the MatchingData passed during matching.
        """
        if clf is None:
            _check_fitted(self)
            clf = self.best_model

        if matching_data is None:
            matching_data = self.matching_data

        X, y = self._preprocess_data_for_sklearn(matching_data)
        propensity_pool = clf.predict_proba(X[y == self.POOL, :])[:, 1]
        propensity_target = clf.predict_proba(X[y == self.TARGET, :])[:, 1]

        return propensity_pool, propensity_target


def propensity_score_match_greedy(ps_pool, ps_target, caliper=None, clip=True):
    # copy array, since we will be changing it and don't want to impact caller
    ps_pool = copy.copy(ps_pool)
    if caliper is not None and clip:
        ps_pool[ps_pool > 1 - caliper] = -np.inf

    pool_matches = []
    target_matches = []
    for i, ps_pat in enumerate(ps_target):
        # find closest patient in ps_pool based on propensity score
        scores = np.abs(ps_pool - ps_pat)
        if caliper is not None:
            if all(scores > caliper) or (clip and not caliper <= ps_pat <= 1 - caliper):
                continue
        match = np.argmin(scores)
        pool_matches.append(match)
        target_matches.append(i)

        # set the propensity score for the matched patient to negative infinity
        # to ensure the patient is never matched again
        ps_pool[match] = -np.inf

    return np.array(pool_matches), np.array(target_matches)


def propensity_score_match_greedy_prio(ps_pool, ps_target, caliper=None):
    """
    An alternative greedy approach in which we still do greedy matching but we
    sort patients by how "close" they are to the pool PS distribution. Patients
    that are furthest from the pool distribution are matched first.
    """
    pool_matches, target_matches = [], []
    ps_broad = np.abs(np.subtract.outer(ps_pool, ps_target))
    order = ps_broad.mean(axis=0).argsort()[::-1]
    pool_matches, target_matches = propensity_score_match(
        ps_pool, ps_target[order], method="greedy", caliper=caliper
    )
    target_matches = order[target_matches]
    return np.array(pool_matches), np.array(target_matches)


def propensity_score_match_linear_sum_assignment(ps_pool, ps_target):
    scores = np.ones((len(ps_pool), len(ps_target)), dtype=np.float32)
    scores = np.abs(scores * np.expand_dims(ps_pool, -1) - ps_target)
    pool_matches, target_matches = optimize.linear_sum_assignment(scores)
    return pool_matches, target_matches


def propensity_score_match(ps_pool, ps_target, method="greedy", caliper=None):
    """
    Return the set of indices to use for matching an RCT to RWD based on
    propensity scores. Inputs are propensity scores, output is index to real
    world data array.

    ps_pool: np.array
        propensity scores of the population to be matched
    ps_target: np.array
        propensity scores of the reference population (usually the RCT
        population)
    method: string
        method to use for matching, can be one of the following:
           greedy -- match each RCT patient to its nearest match by propensity
           greedy_prio -- greedy but use a heuristic to sort patients and match
           the hardest patients first
           score linear_sum_assignment -- return a
           globally optimum pairing that minimizes the absolute sum
                of differences between the RCT patients and the real world
                patients

    :return: Tuple of numpy arrays pool_matches, target_matches whose elements
        indicate the indices of matched patients in the pool and target
        populations respectively.
    """
    ps_pool, ps_target = copy.deepcopy(ps_pool), copy.deepcopy(ps_target)
    if method == "greedy":
        pool_matches, target_matches = propensity_score_match_greedy(
            ps_pool, ps_target, caliper=caliper
        )
    elif method == "greedy_prio":
        pool_matches, target_matches = propensity_score_match_greedy_prio(
            ps_pool, ps_target, caliper=caliper
        )

    elif method == "linear_sum_assignment":
        if caliper is not None:
            logger.error("Caliper is not supported for linear_sum_assignment!")
        pool_matches, target_matches = propensity_score_match_linear_sum_assignment(
            ps_pool, ps_target
        )

    else:
        raise NotImplementedError(f"Matching method {method} not currently supported.")

    return pool_matches, target_matches


def plot_propensity_score_match_distributions(matcher: PropensityScoreMatcher):
    """
    Plot histograms of the estimated propensity score for pool and target
    populations pre- and post-matching.

    :param matcher: Fitted PropensityScoreMatcher model.
    """
    _check_fitted(matcher)
    m = matcher.matching_data
    match = matcher.best_match
    target, pool = split_target_pool(match)
    target_name, pool_name = (
        target[match.population_col].unique()[0],
        pool[match.population_col].unique()[0],
    )

    ps_pool_match, ps_target_match = matcher.get_propensity_score(matching_data=match)
    ps_pool, ps_target = matcher.get_propensity_score(matching_data=m)

    data = pd.concat(
        [
            pd.DataFrame.from_dict(
                {
                    "propensity": ps_pool_match,
                    "matched": True,
                    "population": pool_name,
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "propensity": ps_target_match,
                    "matched": True,
                    "population": target_name,
                }
            ),
            pd.DataFrame.from_dict(
                {"propensity": ps_pool, "matched": False, "population": pool_name}
            ),
            pd.DataFrame.from_dict(
                {"propensity": ps_target, "matched": False, "population": target_name}
            ),
        ]
    )

    g = sns.FacetGrid(data=data, col="matched", height=4, xlim=[0, 1])
    g.map_dataframe(
        sns.histplot,
        bins=np.linspace(0, 1, 25),
        x="propensity",
        hue="population",
        hue_order=[pool_name, target_name],
        alpha=0.5,
        common_norm=False,
        stat="probability",
    )
    [ax.grid(True) for axes in g.axes for ax in axes]

    legend_patches = [
        matplotlib.patches.Patch(color=sns.color_palette()[0], label=pool_name),
        matplotlib.patches.Patch(color=sns.color_palette()[1], label=target_name),
    ]
    matplotlib.pyplot.legend(handles=legend_patches)

    return g


def plot_propensity_score_match_pairs(matcher: PropensityScoreMatcher):
    """
    Plot scatterplot of pool-target pairs formed by propensity score matching.

    :param matcher: Fitted PropensityScoreMatcher model.
    """
    _check_fitted(matcher)
    ps_pool, ps_target = matcher.get_propensity_score()
    pool_matches, target_matches = propensity_score_match(
        ps_pool, ps_target, method=matcher.method, caliper=matcher.caliper
    )
    plt.figure()
    plt.scatter(
        [ps_target[i] for i in target_matches], [ps_pool[i] for i in pool_matches]
    )
    plt.grid()
    plt.xlabel("PS Target")
    plt.ylabel("PS Pool (Matched)")
    plt.plot([0, 1], [0, 1])
    return plt.gcf()



from pybalance.visualization.distributions import _debin_features


def test_debin():
    effective_features = ["x_1", "x_123", "z_10", "z_0", "y_12"]
    input_output_column_mapping = {
        "x": ["x_1", "x_123"],
        "y": ["y_12"],
        "z": ["z_10", "z_0"],
    }
    indices = _debin_features(effective_features, input_output_column_mapping)
    assert indices["x"] == [0, 1]
    assert indices["y"] == [4]
    assert indices["z"] == [2, 3]



from typing import Dict, Union, Optional
from collections import defaultdict

import pandas as pd
import numpy as np
import torch

from pybalance.utils import (
    MatchingData,
    split_target_pool,
    BaseMatchingPreprocessor,
    DecisionTreeEncoder,
    StandardMatchingPreprocessor,
    BetaXPreprocessor,
    GammaPreprocessor,
    GammaXPreprocessor,
)

import logging

logger = logging.getLogger(__name__)


def map_input_output_weights(
    preprocessor: BaseMatchingPreprocessor,
    weights: Optional[Dict[str, float]] = None,
) -> Dict[str, float]:
    """
    Map weights on input features to weights at the output of a given
    preprocessor. This mapping is only possible if the preprocessor defines
    get_feature_names_out(). Weights from the input variable are passed along to
    all output variables. For instance, if "age" is an input variable and gets a
    weight of 10, then each age bin feature will have a weight of 10.

    We also considered "diluting" weights such that the total initial weight is
    spread across the bins. We found, however, that this gives unsatisfactory
    results, since requirements for matching success are usually stated in terms
    of unweighted bins (e.g. |SMD| < 0.1). If for instance age is an important
    feature, then so should any feature constructed from age.
    """
    if weights is None:
        weights = {}
    input_weights = defaultdict(lambda: 1)
    input_weights.update(weights)
    output_weights = defaultdict(lambda: 1)

    for input_col in preprocessor.input_headers["all"]:
        output_cols = preprocessor.get_feature_names_out(input_col)
        weight_in = input_weights[input_col]

        # Pass weight along to derived features
        weight_out = weight_in
        for output_col in output_cols:
            # if we've seen this column before, something is wrong;
            # an output column should correspond to exactly one input column
            assert output_col not in output_weights.keys()
            output_weights[output_col] = weight_out

    assert set(output_weights.keys()) == set(preprocessor.output_headers["all"])

    return output_weights


def reshape_output(f):
    def _f(*args, **kwargs):
        output = f(*args, **kwargs)
        if len(output) == 1:
            return output[0]
        return output

    return _f


class BaseBalanceCalculator:
    """
    BaseBalanceCalculator is the low-level interface to calculating balance.
    BaseBalanceCalculator can be used with any preprocessor defined as a
    subclass of BaseMatchingPreprocessor. BaseBalanceCalculator implements
    matrix calculations in pytorch to allow for GPU acceleration.

    BaseBalanceCalculator performs two main tasks:

      (1) Computes a per-feature-loss based on the output features of the given
          preprocessor and

      (2) Aggregates the per-feature-loss into a single value for the loss.

    Furthermore, the calculator can compute the loss for many populations at a
    time.

    :matching_data: Input matching data to be used for distance calculations.
        Must contain exactly two populations. The smaller population is used as
        a reference population. Calls to distance() compute the distance to this
        reference population.

    :preprocessor: Preprocessor to use for per-feature-loss calculation. The
        per-feature-loss is, up to some normalizations, the mean difference in
        the features at the output of the preprocessor.

    :feature_weights: How to weight features in aggregation of per-feature-loss.

    :order: Exponent to use in combining per-feature-loss into an aggregate
        loss. Total loss is sum(feature_weight * feature_loss**order)**(1/order).

    :param standardize_difference: Whether to use the absolute standardized mean
        difference for the per-feature loss (otherwise uses absolute mean
        difference).

    :device: Name of device to use for matrix computations. By default, will use
        GPU if a GPU is found on the system.
    """

    name = "base"

    def __init__(
        self,
        matching_data: MatchingData,
        preprocessor: BaseMatchingPreprocessor,
        feature_weights: Optional[Dict[str, float]] = None,
        order: float = 1,
        standardize_difference: bool = True,
        device: Optional[str] = None,
    ):
        self.order = order
        self.standardize_difference = standardize_difference
        self.device = self._get_device(device)
        self.preprocessor = preprocessor
        self.preprocessor.fit(matching_data)
        self.matching_data = matching_data
        target, pool = split_target_pool(matching_data)
        self.target = self._preprocess(target)
        self.pool = self._preprocess(pool)
        self._set_feature_weights(feature_weights)

        self.target_mean = torch.mean(self.target, 0, True).to(self.device)
        self.target_std = torch.std(self.target, 0, keepdim=True).to(self.device)

        # Zero variances are bad and can lead to infinite loss.
        if any((self.target_std == 0)[0]):
            bad_columns = [
                self.preprocessor.output_headers["all"][j]
                for j, std in enumerate(self.target_std[0])
                if std == 0
            ]
            logger.warning(
                f'Detected constant feature(s) in target population: {",".join(bad_columns)}.'
            )

        pool_std = torch.std(self.pool, 0, keepdim=True).to(self.device)
        # Zero variances are bad and can lead to infinite loss.
        if any((pool_std == 0)[0]):
            bad_columns = [
                self.preprocessor.output_headers["all"][j]
                for j, std in enumerate(pool_std[0])
                if std == 0
            ]
            logger.warning(
                f'Detected constant feature(s) in pool population: {",".join(bad_columns)}.'
            )

    def _set_feature_weights(self, feature_weights):
        if feature_weights is None:
            feature_weights = {}
            try:
                feature_weights = map_input_output_weights(
                    self.preprocessor, feature_weights
                )
            except NotImplementedError:
                # User has not passed weights, so we can assume that
                # NotImplementedError can be ignored. Note that not passing
                # weights does not mean weights will always be equal!! Equal
                # weights on the input get diluted on the output, depending on
                # how many output features an input feature is mapped to. In the
                # current case, there simply is no mapping available, so we must
                # put equal weights on the output feature space.
                feature_weights = defaultdict(lambda: 1)
        else:
            # User has explicitly passed weights, so NotImplementedError should
            # not be handled
            feature_weights = map_input_output_weights(
                self.preprocessor, feature_weights
            )

        feature_weights = np.array(
            [feature_weights[c] for c in self.preprocessor.output_headers["all"]]
        )
        feature_weights = feature_weights / sum(feature_weights)
        self.feature_weights = torch.tensor(feature_weights).to(self.device)

    def _get_device(self, device):
        if device is None:
            if torch.cuda.is_available():
                device = "cuda:0"
            else:
                device = "cpu"
        return torch.device(device)

    def _preprocess(self, data: Union[pd.DataFrame, MatchingData]):
        if isinstance(data, pd.DataFrame):
            data = MatchingData(
                data=data,
                headers=self.matching_data.headers,
                population_col=self.matching_data.population_col,
            )
        data = self.preprocessor.transform(data)
        data = data[data.headers["all"]].values
        # It's best to store the feature data on the GPU to avoid moving it back
        # and forth. Even for large populations, this will be cheap (e.g.
        # 500,000 patients with 100 features = less than 1GB). Be careful,
        # however: when slicing for the candidate populations, this can blow up
        # the memory footprint. Use the BatchedBalanceCalculator to stay within
        # memory limits.
        data = data.astype(np.float32)
        data = torch.tensor(
            data, dtype=torch.float32, device=self.device, requires_grad=False
        )
        return data

    def _fetch_pool_features(self, candidate_populations):
        if isinstance(candidate_populations, pd.DataFrame):
            features = self._preprocess(candidate_populations)
        else:
            if not isinstance(candidate_populations, torch.Tensor):
                candidate_populations = torch.tensor(
                    candidate_populations, device=self.device, requires_grad=False
                )
            features = self.pool[candidate_populations]

        # features has shape n_candidate_populations x n_patients x n_features
        # note that using the reshape() method works for both numpy and pytorch
        if len(features.shape) == 2:
            features = features.reshape(1, *features.shape)

        return features

    def _finalize_batches(self, batches):
        """
        Combine a list (batches) of batched distances calculations into a single
        pytorch tensor.
        """
        return torch.hstack(batches)

    def _to_array(self, candidate_populations):
        return torch.tensor(candidate_populations, device=self.device)

    def _to_list(self, candidate_populations):
        return candidate_populations.cpu().detach().numpy().tolist()

    @reshape_output
    def distance(
        self, candidate_populations: Union[pd.DataFrame, torch.Tensor, np.ndarray]
    ) -> torch.Tensor:
        """
        Compute overall distance (aka "mismatch", aka "loss") between input
        candidate populations. The per-feature loss is aggregated using a vector
        norm specified by order specified in __init__().

        :candidate_populations: Populations for which to compute the mismatch.
            Input can be either a single pandas dataframe containing all the
            required feature data or a 2-dimensional integer array whose entries
            are the indices of patients in the pool. In the latter case, the
            array should be an array of shape n_candidate_populations x
            candidate_population_size that indexes the patient pool (passed
            during __init__()). The first dimension may be omitted if only one
            candidate population is present.
        """
        per_feature_loss = self.per_feature_loss(candidate_populations)
        # Since feature_weights sum to 1, sum is actually a weighted mean
        return torch.sum(
            self.feature_weights * torch.abs(per_feature_loss) ** self.order, dim=1
        ) ** (1.0 / self.order)

    def balance(self, candidate_populations):
        return -self.distance(candidate_populations)

    def per_feature_loss(
        self, candidate_populations: Union[pd.DataFrame, torch.Tensor, np.ndarray]
    ) -> torch.Tensor:
        """
        Compute mismatch (aka "distance", aka "loss") on a per-feature basis for
        a set of candidate populations.
        """
        pool = self._fetch_pool_features(candidate_populations)

        # NaNs can arise when the pool and target both have zero variance. If
        # they have zero variance and are the same value (e.g. pool all 0s and
        # target all 0s), then this should be anyway zero loss. As long as the
        # norm is non-zero, we are good. If they have zero variance and are
        # different values (pool all 1s and target all 0s), then this should
        # represent a large loss. Technically, the loss should be infinite in
        # that case, but infinities are annoying so we just set the norm to a
        # small positive number. The user will get warned in the call to
        # __init__() of the possibility of this arising.
        norm = 1
        if self.standardize_difference:
            norm *= torch.sqrt(pool.std(axis=1) ** 2 + self.target_std**2) + 1e-6

        return torch.nan_to_num((pool.mean(axis=1) - self.target_mean) / norm)


class BetaBalance(BaseBalanceCalculator):
    """
    Convenience interface to BaseBalanceCalculator to computes the distance
    between populations as the mean standardized mean difference. Uses
    StandardMatchingPreprocessor as the preprocessor.
    """

    name = "beta"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        drop: bool = "first",
        standardize_difference: bool = True,
    ):
        preprocessor = StandardMatchingPreprocessor(drop=drop)
        super(BetaBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=preprocessor,
            feature_weights=feature_weights,
            order=1,
            standardize_difference=standardize_difference,
            device=device,
        )


class BetaSquaredBalance(BaseBalanceCalculator):
    """
    Same as BetaBalance, except that per-feature balances are averaged in a
    mean square fashion.
    """

    name = "beta_squared"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        drop: bool = "first",
        standardize_difference: bool = True,
    ):
        preprocessor = StandardMatchingPreprocessor(drop=drop)
        super(BetaSquaredBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=preprocessor,
            feature_weights=feature_weights,
            order=2,
            standardize_difference=standardize_difference,
            device=device,
        )


class BetaXBalance(BaseBalanceCalculator):
    """
    Convenience interface to BaseBalanceCalculator to compute the balance
    between two populations by computing the standardized mean difference,
    including cross terms. See BetaXPreprocessor for description of
    preprocessing options.
    """

    name = "beta_x"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        drop: str = "first",
        standardize_difference: bool = True,
        max_cross_terms="auto",
    ):
        pp_x = BetaXPreprocessor(
            drop=drop,
            max_cross_terms=max_cross_terms,
        )
        super(BetaXBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=pp_x,
            order=1,
            standardize_difference=standardize_difference,
            device=device,
        )


class BetaXSquaredBalance(BaseBalanceCalculator):
    """
    Same as BetaXBalance, except that per-feature balances are averages in a
    mean square fashion.
    """

    name = "beta_x_squared"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        drop: str = "first",
        standardize_difference: bool = True,
        max_cross_terms="auto",
    ):
        pp_x = BetaXPreprocessor(
            drop=drop,
            max_cross_terms=max_cross_terms,
        )
        super(BetaXSquaredBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=pp_x,
            order=2,
            standardize_difference=standardize_difference,
            device=device,
        )


class BetaMaxBalance(BaseBalanceCalculator):
    """
    Same as BetaBalance, except the worst-matched feature determines the loss.
    This class is provided as a convenience, since this balance metric is often
    a criterion used to determine if matching is "sufficiently good". However,
    be aware that using this balance metric as an optimization objective with
    the various matchers can lead unwanted behavior, since if improvements in
    the worst-matched feature are not possible, there is no signal from the
    balance function to improve any other the other features.
    """

    name = "beta_max"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        drop: bool = "first",
        standardize_difference: bool = True,
    ):
        preprocessor = StandardMatchingPreprocessor(drop=drop)
        super(BetaMaxBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=preprocessor,
            feature_weights=feature_weights,
            order=1,
            standardize_difference=standardize_difference,
            device=device,
        )

    @reshape_output
    def distance(
        self, candidate_populations: Union[pd.DataFrame, torch.Tensor, np.ndarray]
    ) -> torch.Tensor:
        per_feature_loss = self.per_feature_loss(candidate_populations)

        return torch.max(
            self.feature_weights * torch.abs(per_feature_loss), dim=1
        ).values


class GammaBalance(BaseBalanceCalculator):
    """
    Convenience interface to BaseBalanceCalculator to compute the balance
    between two populations by computing the mean area between their
    one-dimensional marginal distributions. See GammaPreprocessor for
    description of preprocessing options.
    """

    name = "gamma"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        n_bins: int = 5,
        encode: str = "onehot-dense",
        cumulative: bool = True,
        drop: str = "first",
        standardize_difference: bool = True,
    ):
        preprocessor = GammaPreprocessor(
            n_bins=n_bins, encode=encode, cumulative=cumulative, drop=drop
        )
        super(GammaBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=preprocessor,
            feature_weights=feature_weights,
            order=1,
            standardize_difference=standardize_difference,
            device=device,
        )


class GammaSquaredBalance(BaseBalanceCalculator):
    """
    Same as GammaBalance, except that per-feature balances are averages in a
    mean square fashion.
    """

    name = "gamma_squared"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        n_bins: int = 5,
        encode: str = "onehot-dense",
        cumulative: bool = True,
        drop: str = "first",
        standardize_difference: bool = True,
    ):
        preprocessor = GammaPreprocessor(
            n_bins=n_bins, encode=encode, cumulative=cumulative, drop=drop
        )
        super(GammaSquaredBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=preprocessor,
            feature_weights=feature_weights,
            order=2,
            standardize_difference=standardize_difference,
            device=device,
        )


class GammaXBalance(BaseBalanceCalculator):
    """
    Convenience interface to BaseBalanceCalculator to compute the balance
    between two populations by computing the mean area between their
    one-dimensional marginal distributions, including cross terms. See
    GammaXPreprocessor for description of preprocessing options.
    """

    name = "gamma_x"

    def __init__(
        self,
        matching_data: MatchingData,
        feature_weights: Optional[Dict[str, float]] = None,
        device: Optional[str] = None,
        n_bins: int = 5,
        encode: str = "onehot-dense",
        cumulative: bool = True,
        drop: str = "first",
        standardize_difference: bool = True,
        max_cross_terms="auto",
    ):
        pp_x = GammaXPreprocessor(
            n_bins=n_bins,
            encode=encode,
            cumulative=cumulative,
            drop=drop,
            max_cross_terms=max_cross_terms,
        )
        super(GammaXBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=pp_x,
            order=1,
            standardize_difference=standardize_difference,
            device=device,
        )


class GammaXTreeBalance(BaseBalanceCalculator):
    name = "gamma_x_tree"

    def __init__(
        self,
        matching_data,
        keep_original_features=False,
        device=None,
        standardize_difference: bool = True,
        **decision_tree_params,
    ):
        pp_tree = DecisionTreeEncoder(
            keep_original_features=keep_original_features, **decision_tree_params
        )
        super(GammaXTreeBalance, self).__init__(
            matching_data=matching_data,
            preprocessor=pp_tree,
            order=1,
            standardize_difference=standardize_difference,
            device=device,
        )


def _get_batch_size(target_population_size, n_features, max_batch_size_gb=8):
    """
    Get the size of batches for balance calculations such that no batch is
    greater in memory footprint than max_size_gb GB. This calculation is only
    approximate and not guaranteed to respect size requirements. n_features
    should be interpreted as the number of *effective* features (i.e. after
    binning)
    """
    max_batch_size_gb = max_batch_size_gb / 1.25  # add some wiggle room
    size_of_float = (
        4  # Balance calculators are supposed to transform to float datatypes!
    )
    # Technically, the size is given by:
    # batch_size_gb = n_candidate_populations * target_population_size * n_features * size_of_float / 2**30
    # But this value will cancel out in the next step.
    batch_size_gb = target_population_size * n_features * size_of_float / 2**30
    batch_size = int(max_batch_size_gb / batch_size_gb)

    # If you fail this assertion, something is definitely wrong and you'll pay
    # for it downstream. Rather stop you here. Basically, it means you can't
    # even hold 10 candidate populations in memory at a time.
    assert batch_size >= 10
    return batch_size


class BatchedBalanceCaclulator:
    """
    Batch balance calculations to avoid large peak memory usage.
    """

    def __init__(self, balance_calculator, max_batch_size_gb=8):
        self.name = balance_calculator.name
        self.matching_data = balance_calculator.matching_data
        self.preprocessor = balance_calculator.preprocessor
        self.pool = balance_calculator.pool
        self.target = balance_calculator.target
        self.balance_calculator = balance_calculator
        self.max_batch_size_gb = max_batch_size_gb

    def _to_array(self, candidate_populations):
        return self.balance_calculator._to_array(candidate_populations)

    def _to_list(self, candidate_populations):
        return self.balance_calculator._to_list(candidate_populations)

    def distance(self, candidate_populations):
        # If the user passes a pandas DataFrame, the DataFrame is assumed to
        # represent feature data for only one population and no need for
        # batching; just pass on to the base class. Otherwise, assume
        # candidate_populations refers to patient indices.
        if isinstance(candidate_populations, pd.DataFrame):
            return self.balance_calculator.distance(candidate_populations)

        # If the user passes a list, convert it to the underlying backend array
        # type for further processing.
        if isinstance(candidate_populations, list):
            candidate_populations = self.balance_calculator._to_array(
                candidate_populations
            )

        # If array has only one dimension, no need for batching, just pass along
        # to base class.
        if len(candidate_populations.shape) == 1:
            return self.balance_calculator.distance(candidate_populations)

        # Get batch size according to number of size of target population
        batch_size = _get_batch_size(
            self.balance_calculator.target.shape[0],
            self.balance_calculator.target.shape[1],
            self.max_batch_size_gb,
        )

        # From here on, can assume candidate_populations is a 2D array of
        # patient indices with the 0th dimension corresponding to the
        # candidate_population and the 1st dimension corresponding to patient.
        n_remaining = len(candidate_populations)
        distances = []
        j = 0
        while n_remaining > 0:
            N = min(batch_size, n_remaining)
            cps = candidate_populations[j * batch_size : j * batch_size + N, :]
            distances.append(self.balance_calculator.distance(cps))
            n_remaining -= N
            j += 1

        distances = self.balance_calculator._finalize_batches(distances)
        assert len(distances) == len(candidate_populations)
        return distances

    def balance(self, candidate_populations):
        return -self.distance(candidate_populations)


#
# Convenience interface to balance calculators
#
BALANCE_CALCULATORS = {
    BaseBalanceCalculator.name: BaseBalanceCalculator,
    BetaBalance.name: BetaBalance,
    BetaXBalance.name: BetaXBalance,
    BetaXSquaredBalance.name: BetaXSquaredBalance,
    BetaMaxBalance.name: BetaMaxBalance,
    BetaSquaredBalance.name: BetaSquaredBalance,
    GammaBalance.name: GammaBalance,
    GammaSquaredBalance.name: GammaSquaredBalance,
    GammaXTreeBalance.name: GammaXTreeBalance,
    GammaXBalance.name: GammaXBalance,
}


def BalanceCalculator(matching_data, objective="gamma", **kwargs):
    """
    BalanceCalculator provides a convenience interface to balance calculators,
    allowing the user to initialize a balance calculator by name. The
    calculators are initialized with default parameters, but these can be
    overridden by passing the appropriate kwargs.

    :param matching_data: MatchingData instance containing reference to the data
        against which matching metrics will be computed

    :param objective: Name of objective function to be used for computing
        balance. Balance calculators must be implemented in
        utils.balance_calculators.py and registered in the BALANCE_CALCULATORS
        dictionary therein in order to be accessible from this interface.

    :param kwargs: Any additional arguments required to configure the specific
        objective function (e.g. n_bins = 10 for "gamma").
    """
    if objective not in BALANCE_CALCULATORS.keys():
        raise ValueError(
            f"Unknown objective function {objective}. Must be one of {','.join(BALANCE_CALCULATORS.keys())}"
        )
    balance_calculator = BALANCE_CALCULATORS[objective]
    return balance_calculator(matching_data, **kwargs)



import numpy as np
from pybalance.propensity import matcher


def test_greedy_without_replacement():
    rw_ptt = np.array(
        [
            0.1,
            0.2,
            0.3,
            0.4,
        ]
    )
    rct_ptt = np.array([0.5, 0.6, 0.7])
    rw, rct = matcher.propensity_score_match(rw_ptt, rct_ptt, method="greedy")
    assert set(rw) == set([1, 2, 3])


def test_lsa():
    rw_ptt = np.array([0.1, 0.2, 0.3, 0.4])
    rct_ptt = np.array([0.5, 0.6, 0.7])
    rw, rct = matcher.propensity_score_match(
        rw_ptt,
        rct_ptt,
        method="linear_sum_assignment",
    )
    assert set(rw) == set([1, 2, 3])


def test_ps_greedy_w_calliper():
    rw_ptt = np.array([0.1, 0.2, 0.3, 0.0, 0.5])
    rct_ptt = np.array([0.2, 0.2, 0.8])
    rw, rct = matcher.propensity_score_match_greedy(rw_ptt, rct_ptt)
    assert all(rw == np.array([1, 2, 4]))
    assert all(rct == np.array([0, 1, 2]))

    rw, rct = matcher.propensity_score_match_greedy(rw_ptt, rct_ptt, caliper=0.1)
    assert all(rw == np.array([1, 2]))
    assert all(rct == np.array([0, 1]))



from collections import deque
import numpy as np
import pandas as pd
from typing import Union
import time

import logging

logger = logging.getLogger(__name__)

from pybalance.utils.balance_calculators import (
    BalanceCalculator,
    BatchedBalanceCaclulator,
)
from pybalance.utils import (
    MatchingData,
    split_target_pool,
    BaseBalanceCalculator,
)
from pybalance.genetic.initialization import GeneticMatcherInitializer
from pybalance.genetic.logger import BasicLogger
import torch


def _check_fitted(matcher):
    if matcher.best_match is None:
        raise (ValueError, "Matcher has not been fitted!")


def get_global_defaults(n_candidate_populations=5000):
    """
    Get a set of reasonable default values for evolutionary configuration. We
    break parameters into two groups: evolutionary, i.e., those that govern how
    the candidate populations are mixed, and initialization, i.e., those that
    govern the initial set of candidate populations.

    :param n_candidate_populations: Number of candidate populations to evolve.
    """
    #
    # Evolutionary params -- how to mix populations
    #
    config = {
        # Size of candidate population to match against reference data. If not
        # specified, will use the same size as the reference population.
        "candidate_population_size": None,
        # n_candidate_populations = number of candidate populations to simultaneously evolve.
        # n_keep_best = keep top N current best scoring candidate populations
        # n_voting_populations = form new candidate populations based on frequency of patient occurence
        # n_mutation = make individual patient swaps for top candidate populations
        "n_candidate_populations": n_candidate_populations,
        "n_keep_best": int(n_candidate_populations / 4),
        "n_voting_populations": int(n_candidate_populations / 4),
        "n_mutation": int(n_candidate_populations / 4),
        # n_generations = number of generations to evolve the candidate populations.
        # time_limit = time limit (in seconds), checked only at end of every iteration
        "n_generations": 1000,
        "n_iter_no_change": 100,
        "time_limit": None,
        "max_batch_size_gb": 2,
        "seed": 1234,
        "verbose": True,
        "log_every": 5,
    }

    #
    # Initialization params -- how to initialize the candidate populations
    #
    config["initialization"] = {
        "benchmarks": {"propensity": "include"},
        "sampling": {
            "propensity": 1.0,
            "uniform": 1.0,
        },
    }

    return config


class GeneticMatcher:
    """
    Match two populations using a genetic algorithm.

    :param matching_data: MatchingData to be matched. Must contain exactly two
        populations. The larger population will be matched to the smaller.

    :param objective: Matching objective to optimize in hyperparameter search.
        Can be a string referring to any balance calculator known to
        utils.balance_calculators.BalanceCalculator or an instance of
        BaseBalanceCalculator.

    :param params: Configuration params for the genetic matcher. See
        pybalance.genetic.get_global_defaults for a list of options.
    """

    def __init__(
        self,
        matching_data: MatchingData,
        objective: Union[str, BaseBalanceCalculator] = "beta",
        **params,
    ):
        self.matching_data = matching_data
        self.target, self.pool = split_target_pool(matching_data)

        if isinstance(objective, str):
            self.balance_calculator = BalanceCalculator(self.matching_data, objective)
            self.objective = objective
        else:
            self.balance_calculator = objective
            self.objective = self.balance_calculator.name

        params = self._check_params(params)
        self.params = {"objective": self.objective}
        self.set_params(**params)

        self.balance_calculator = BatchedBalanceCaclulator(
            self.balance_calculator, self.max_batch_size_gb
        )

        if torch.cuda.is_available():
            self.device = torch.device("cuda:0")
        else:
            self.device = torch.device("cpu")

        self.logger = BasicLogger(params.pop("log_every"))
        logger.info(self.device)

        self._reset_best_match()

    def _check_params(self, config):
        # set default values for configuration parameters and check sanity
        # of resulting parameter set

        n_candidate_populations = config.setdefault("n_candidate_populations", 1024)
        standard_config = get_global_defaults(n_candidate_populations)
        standard_config.update(config)

        if standard_config["candidate_population_size"] is None:
            standard_config["candidate_population_size"] = len(self.target)
        if standard_config["n_keep_best"] >= standard_config["n_candidate_populations"]:
            raise ValueError(
                "GeneticMatcher cannot train if it keeps all candidate populations. \
                Either select more candidate populations or keep fewer from round to round."
            )
        if standard_config["n_keep_best"] <= 1:
            raise ValueError("n_keep_best must be > 1")
        if standard_config["seed"] is not None:
            np.random.seed(standard_config["seed"])

        return standard_config

    def set_params(self, **kwargs):
        self.params.update(kwargs)
        for kw, val in kwargs.items():
            setattr(self, kw, val)

    def get_params(self):
        return self.params

    def _reset_best_match(self):
        self.best_match = None
        self.best_match_idx = None
        self.best_score = np.inf
        self.balance = None
        self._recent_balance = deque([], maxlen=self.params["n_iter_no_change"])
        self.generation = 0
        self.elapsed_time = 0

    def match(self, seed=None):
        """
        Match populations passed during __init__(). Returns MatchingData
        instance containing the matched pool and target populations.
        """
        t0 = time.time()
        self._init_first_generation(seed)
        stop = self._check_stopping_conditions()
        while not stop:
            self._log()
            self._generate_offspring()
            self.elapsed_time = time.time() - t0
            stop = self._check_stopping_conditions()

        self._log(finalize=True)
        return self.get_best_match()

    def _init_first_generation(self, seed=None):
        initializer = GeneticMatcherInitializer(self)
        candidate_populations = initializer.initialize(
            self.n_candidate_populations, seed=seed
        )
        self.candidate_populations = torch.tensor(np.array(candidate_populations)).to(
            self.device
        )

    def _check_stopping_conditions(self):
        best_match = max(self.balance)
        self._recent_balance.append(best_match)
        if best_match == 0:
            logger.info("Optimal solution found! Stopping")
            return True
        if (len(self._recent_balance) == self._recent_balance.maxlen) and (
            best_match == self._recent_balance[0]
        ):
            logger.info(
                f"No improvement in last {self._recent_balance.maxlen} iterations. Stopping."
            )
            return True
        if self.generation == self.n_generations:
            logger.info("Reached maximum number of generations. Stopping.")
            return True
        if self.time_limit is not None and self.elapsed_time > self.time_limit:
            logger.info("Time limit exceeded. Stopping.")
            return True
        return False

    def _log(self, finalize=False):
        if self.logger is not None:
            self.logger.on_generation_end(self)
            if finalize:
                self.logger.on_matching_end(self)

    def _generate_offspring(self):
        """
        Create a new set of candidate_populations by preferentially mating
        fitter i.e. more similar to target candidate_populations. Higher balance
        is more likely to mate
        """
        self.generation += 1

        # keep the best N groups so as to not regress
        # Note that higher balance values are better, so we take the candidate
        # populations from the end of the list
        # FIXME n_keep_best MUST BE AT LEAST ONE OR IT WILL TAKE EVERYTHING!!!
        # balance = torch.tensor(self.balance).to(self.device)
        idxs_best_n_matches = torch.argsort(self.balance)[-self.n_keep_best :]
        offspring = self.candidate_populations[idxs_best_n_matches, :]

        # make individual patient swaps
        if self.n_mutation:
            offspring = torch.vstack(
                (offspring, self.generate_mutated_populations(N=self.n_mutation))
            )

        # add candidate populations formed from the entire pool of candidate populations
        # according to frequency of occurrence
        if self.n_voting_populations:
            offspring = torch.vstack(
                (offspring, self.generate_voting_populations(self.n_voting_populations))
            )

        # perform random mating
        n_remaining = self.n_candidate_populations - offspring.shape[0]
        if n_remaining:
            offspring = torch.vstack(
                (offspring, self.generate_mating_populations(N=n_remaining))
            )

        self.candidate_populations = offspring

    def generate_mating_populations(self, N=None, n_way_mating=2):
        # get the mating probabilities for all candidate populations
        # basically equivalent to rankdata which the cpu version uses
        # note 1.0 is needed to convert to float and avoid dropping the worst candpop.
        p = 1.0 + torch.argsort(self.balance)

        # randomly select groups of candidate populations to mix
        weights = p.repeat(N, 1)
        mating_populations = torch.multinomial(
            weights, num_samples=n_way_mating, replacement=False
        )
        mated_populations = self.candidate_populations[mating_populations].reshape(
            N, n_way_mating * self.candidate_population_size
        )

        # shuffle in the last (patient) dimension
        # permutes all cols together but that's ok since the rows are uncorrelated!
        perm = torch.randperm(2 * self.candidate_population_size, device=self.device)
        mated_populations = mated_populations[:, perm]
        mated_populations = torch.vstack(
            [
                torch.unique(t)[: self.candidate_population_size]
                for t in torch.unbind(mated_populations)
            ]
        )

        return mated_populations

    def generate_voting_populations(self, N=None):
        # count patient frequency within candidate populations. this unique
        # function doesn't seem to scale (in terms of peak memory usage) so well
        # when you have a lot of large candidate populations, so here we just
        # estimate the counts based on a sample of the candidate populations
        max_candidate_populations = 1024
        sample_populations = torch.randperm(len(self.candidate_populations))
        patients, counts = torch.unique(
            self.candidate_populations[
                sample_populations[:max_candidate_populations], :
            ],
            return_counts=True,
        )
        voting_probabilities = counts / counts.sum()

        candidate_populations = torch.empty(
            size=(0, self.candidate_population_size),
            device=self.device,
            dtype=self.candidate_populations.dtype,
        )
        while len(candidate_populations) < N:
            # FIXME calculate the max _N based on batch size. I can't quite
            # figure out what the memory requirements of the operation below are
            # so I don't know how to do that. For now, just hard code to a nice
            # power of two
            this_N = min(max_candidate_populations, N - len(candidate_populations))
            weights = voting_probabilities.repeat(this_N, 1)
            this_candidate_populations = torch.multinomial(
                weights, num_samples=self.candidate_population_size, replacement=False
            )
            candidate_populations = torch.vstack(
                [candidate_populations, this_candidate_populations]
            )

        return candidate_populations

    def generate_mutated_populations(self, N=None, n_swap=1):
        mutaters = torch.argsort(self.balance)[-N:]
        mutated_populations = self.candidate_populations[
            mutaters
        ]  # N x candidate_population_size
        torch.ones(
            n_swap,
        )

        random_pool_patients = torch.randperm(len(self.pool), device=self.device)[
            :N
        ].reshape(N, 1)
        mutated_populations = torch.hstack([random_pool_patients, mutated_populations])

        # would love to avoid this for loop but I don't see how. For some
        # reason, the shuffling is absolutely needed. Such is life. It's still
        # faster that numpy, even on a CPU.
        out = []
        for t in torch.unbind(mutated_populations):
            t = torch.unique(t)
            t = t[torch.randperm(len(t), device=self.device)][
                : self.candidate_population_size
            ]
            out.append(t)

        mutated_populations = torch.vstack(out)

        return mutated_populations

    def _calculate_balance(self):
        self.balance = self.balance_calculator.balance(self.candidate_populations)

    @property
    def candidate_populations(self):
        return self._candidate_populations

    @candidate_populations.setter
    def candidate_populations(self, value):
        self._candidate_populations = value
        # To avoid having the balance array get out of sync with the candidate
        # populations, always recompute balance whenever updating the candidate
        # populations
        self._calculate_balance()

    def get_best_match_idxs(self, balance_calculator=None):
        if balance_calculator is not None:
            balance = balance_calculator.balance(self.candidate_populations)
        else:
            balance = self.balance
        idx_best_match = balance.argmax()
        return self.candidate_populations[idx_best_match, :]

    def get_best_match(self, balance_calculator=None):
        best_match_patient_idxs = (
            self.get_best_match_idxs(balance_calculator).cpu().numpy()
        )
        pool = self.pool.iloc[best_match_patient_idxs]
        target = self.target
        match = MatchingData(
            data=pd.concat([target, pool]),
            headers=self.matching_data.headers,
            population_col=self.matching_data.population_col,
        )

        return match



from .distributions import (
    plot_numeric_features,
    plot_categoric_features,
    plot_binary_features,
    plot_per_feature_loss,
    plot_joint_numeric_distributions,
    plot_joint_numeric_categoric_distributions,
)



def require_fitted(f):
    """
    Use as a decorator on all methods that require a fitted instance (e.g. the
    transform method in preprocessors).
    """

    def _f(self, *args, **kwargs):
        if not self.is_fitted:
            raise RuntimeError(
                f"Method {f.__name__} cannot be called before calling .fit()!"
            )
        return f(self, *args, **kwargs)

    return _f



"""
Some helpful functions for plotting distribution.
"""
from collections import defaultdict
from typing import List, Optional
import itertools

import matplotlib

matplotlib.use("agg")
from matplotlib import pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd

from pybalance import MatchingData, BaseBalanceCalculator, split_target_pool

import logging

logger = logging.getLogger(__name__)


def _get_default_hue_order(matching_data: MatchingData) -> List[str]:
    return sorted(matching_data.populations)


def _get_reference_population(matching_data: MatchingData) -> str:
    # Try to make a reasonable choice for what to use as a reference calculation
    # when computing differences.
    try:
        # If two populations defined, use target
        target, pool = split_target_pool(matching_data)
        reference_population = target[matching_data.population_col].unique()[0]
    except ValueError:
        try:
            # If something called 'target' exists, use it
            matching_data.get_population("target")
            reference_population = "target"
        except KeyError:
            # Else pick one of the smallest populations
            reference_population = (
                matching_data.counts()
                .reset_index()
                .sort_values([matching_data.population_col, "N"])
                .head(1)[matching_data.population_col]
                .values[0]
            )
    return reference_population


def _debin_features(effective_features, input_output_column_mapping):
    # Map original features to effective features
    indices = defaultdict(list)
    original_features = input_output_column_mapping.keys()
    for feature in original_features:
        for j, new_feature in enumerate(effective_features):
            if new_feature in input_output_column_mapping[feature]:
                indices[feature].append(j)
    if not len(sum(list(indices.values()), start=[])) == len(effective_features):
        raise ValueError("debinning not possible, try reruning with debin=False")
    return indices


def _plot_1d_marginals(matching_data, headers, col_wrap, height, **plot_params):
    # Set up figure of correct size and shape.
    ncols = col_wrap
    nrows = 1 + (len(headers) - 1) // ncols
    fig = plt.figure(figsize=(height * ncols, 3 * height / 4 * nrows))

    # PLOT!
    data = matching_data.data
    for j, col in enumerate(headers):
        ax = plt.subplot(nrows, ncols, j + 1)
        sns.histplot(data=data, x=col, **plot_params, ax=ax)
        ax.grid(True)

    # Align axes
    if plot_params["cumulative"]:
        ymax = 1
    else:
        ymax = max([ax.get_ylim()[1] for ax in fig.axes])
    [ax.set_ylim(0, ymax) for ax in fig.axes]

    return fig


def plot_categoric_features(
    matching_data: MatchingData,
    col_wrap: int = 2,
    height: float = 6,
    include_binary=True,
    **plot_params,
) -> plt.Figure:
    """
    Plot the one-dimensional marginal distributions for all categoric features
    and all treatment groups found in matching_data. Extra keyword arguments are
    passed to seaborn.histplot and override defaults.
    """
    # Set up default plotting params for categoric varaibles.
    default_params = {
        "hue": matching_data.population_col,
        "hue_order": _get_default_hue_order(matching_data),
        "stat": "probability",
        "cumulative": False,
        "common_norm": False,
        "discrete": True,
        "multiple": "dodge",
        "shrink": 0.8,
    }

    # overrides default settings (including hue_order, if supplied)
    default_params.update(plot_params)

    # Determine which covariates to plot.
    headers = matching_data.headers["categoric"]
    if not include_binary:
        headers = [c for c in headers if matching_data[c].nunique() > 2]

    # PLOT!
    fig = _plot_1d_marginals(matching_data, headers, col_wrap, height, **default_params)
    [
        fig.axes[j].set_xticks(matching_data[col].unique())
        for j, col in enumerate(headers)
    ]

    return fig


def plot_numeric_features(
    matching_data: MatchingData, col_wrap: int = 2, height: float = 6, **plot_params
) -> plt.Figure:
    """
    Plot the one-dimensional marginal distributions for all numerical features
    and all treatment groups found in matching_data. Extra keyword arguments are
    passed to seaborn.histplot and override defaults.
    """
    # Set up default plotting params for numeric varaibles.
    default_params = {
        "hue": matching_data.population_col,
        "hue_order": _get_default_hue_order(matching_data),
        "stat": "probability",
        "cumulative": True,
        "common_norm": False,
        "discrete": False,
        "element": "step",
        "bins": 500,
        "linewidth": 3,
        "fill": False,
    }

    # overrides default settings (including hue_order, if supplied)
    default_params.update(plot_params)

    # Determine which covariates to plot.
    headers = matching_data.headers["numeric"]

    # PLOT!
    fig = _plot_1d_marginals(matching_data, headers, col_wrap, height, **default_params)

    return fig


def plot_binary_features(
    matching_data: MatchingData,
    max_features: int = 25,
    include_only: Optional[List[str]] = None,
    orient_horizontal: bool = False,
    standardize_difference: bool = False,
    reference_population: Optional[str] = None,
    **plot_params,
) -> plt.Figure:
    """
    Plot all binary features for all treatment groups found in matching_data.
    Additional keyword arguments are passed to sns.barplot and override default.

    :param matching_data: MatchingData instance containing at least a pool and
        target population.

    :param max_features: Max number of features to show in plot, in case there
        are a lot of binary features. Features are sorted in descending order by
        the initial mismatch between pool and target. The top max_features will
        be shown.

    :param include_only: List of features to consider for plotting. Otherwise,
        all binary features are plotted.

    :param orient_horizontal: If True, orient features along the x-axis.
        Otherwise, features will be along the y-axis.

    :param standardize_difference: Whether to use the absolute standardized mean
        difference for the differences plot (otherwise plots absolute mean
        difference).

    :param reference_population: Name of population in matching_data against
        which other populations should be compared. If not supplied, will use
        the smaller population as the reference population.

    :param plot_params: Parameters passed on to seaborn routines.
    """
    if reference_population is None:
        reference_population = _get_reference_population(matching_data)

    default_params = {
        "hue": matching_data.population_col,
        "hue_order": _get_default_hue_order(matching_data),
    }
    default_params.update(plot_params)

    binary_cols = [
        c
        for c in matching_data.headers["categoric"]
        if matching_data.data[c].nunique() == 2
    ]
    binary_cols = [c for c in binary_cols if include_only is None or c in include_only]

    if len(binary_cols) == 0:
        logger.warning("No binary features found!!")
        return None

    data = matching_data.copy().data
    data.loc[:, binary_cols] = data[binary_cols].rank(method="dense") - 1

    # Frequencies
    frequencies = (
        data.groupby(matching_data.population_col)[binary_cols].mean().T.reset_index()
    )
    frequencies = pd.melt(
        frequencies, id_vars=["index"], value_vars=default_params["hue_order"]
    )

    # Differences
    target_values = frequencies[
        frequencies[matching_data.population_col] == reference_population
    ]
    frequencies = frequencies.merge(target_values, suffixes=["", "_target"], on="index")
    frequencies.loc[:, "difference"] = np.abs(
        frequencies["value"] - frequencies["value_target"]
    )
    if standardize_difference:
        variance = frequencies["value"] * (1 - frequencies["value"]) + frequencies[
            "value_target"
        ] * (1 - frequencies["value_target"])
        frequencies.loc[:, "difference"] = frequencies.loc[:, "difference"] / np.sqrt(
            variance
        )
        difference_label = "Standard Difference"
    else:
        difference_label = "Abs Difference"

    # Restrict to top features
    frequencies = frequencies.sort_values(
        [matching_data.population_col, "difference"], ascending=[False, False]
    )
    features = (
        frequencies[frequencies[matching_data.population_col] != reference_population][
            ["index"]
        ]
        .drop_duplicates()
        .head(max_features)["index"]
        .values
    )
    frequencies = frequencies[frequencies["index"].isin(features)]

    # Sort features by mismatch in the pool
    pool_frequencies = frequencies[
        frequencies[matching_data.population_col] != reference_population
    ]
    non_pool_frequencies = frequencies[
        frequencies[matching_data.population_col] == reference_population
    ]
    frequencies = pd.concat([pool_frequencies, non_pool_frequencies])
    plt.rc("legend", fontsize=14)

    if orient_horizontal:
        fig, axes = plt.subplots(
            nrows=2,
            ncols=1,
            figsize=(len(pool_frequencies) / 2, 16),
            gridspec_kw={"height_ratios": [1, 3]},
        )

        plt.subplot(2, 1, 1)
        sns.barplot(data=frequencies, y="difference", x="index", **default_params)
        xticks, labels = plt.xticks()
        plt.gca().set_xticks(xticks + 0.5, minor=False)
        plt.gca().set_xticks(xticks, minor=True)
        plt.gca().set_xticklabels([""] * len(labels), minor=True, rotation=90)
        plt.grid(True)
        plt.axhline(
            y=0.1,
            xmin=xticks.min(),
            xmax=xticks.max(),
            c="k",
            lw=2.5,
            zorder=10000,
            linestyle="--",
        )
        plt.ylim([0, 0.25])
        plt.ylabel("Abs Difference", fontsize=18)
        plt.gca().get_legend().remove()
        plt.xlabel("")

        plt.subplot(2, 1, 2)
        sns.barplot(data=frequencies, y="value", x="index", **default_params)
        xticks, labels = plt.xticks()
        plt.gca().set_xticks(xticks + 0.5, minor=False)
        plt.gca().set_xticks(xticks, minor=True)
        plt.gca().set_xticklabels(
            labels, minor=True, rotation=45, ha="right", fontsize=16
        )
        plt.grid(True)
        plt.ylabel("Frequency", fontsize=18)
        plt.xlabel("Feature")

    else:
        fig, axes = plt.subplots(
            nrows=1,
            ncols=2,
            figsize=(16, len(pool_frequencies) / 2),
            gridspec_kw={"width_ratios": [3, 1]},
        )

        plt.subplot(1, 2, 2)
        sns.barplot(data=frequencies, x="difference", y="index", **default_params)
        ticks, labels = plt.yticks()
        plt.gca().set_yticks(ticks + 0.5, minor=False)
        plt.gca().set_yticks(ticks, minor=True)
        plt.gca().set_yticklabels([""] * len(labels), minor=True)
        plt.grid(True)
        plt.axvline(
            x=0.1, ymin=ticks.min(), ymax=ticks.max(), c="k", lw=2.5, linestyle="--"
        )
        plt.xlim([0, 0.25])
        plt.xlabel(difference_label, fontsize=18)
        plt.gca().get_legend().remove()
        plt.ylabel("")

        plt.subplot(1, 2, 1)
        sns.barplot(data=frequencies, x="value", y="index", **default_params)
        ticks, labels = plt.yticks()
        plt.gca().set_yticks(ticks + 0.5, minor=False)
        plt.gca().set_yticks(ticks, minor=True)
        plt.gca().set_yticklabels(labels, minor=True, fontsize=16)
        plt.grid(True)
        plt.xlabel("Frequency", fontsize=18)
        plt.ylabel("Feature")

    plt.tight_layout()
    return fig


def plot_per_feature_loss(
    matching_data: MatchingData,
    balance_calculator: BaseBalanceCalculator,
    reference_population: Optional[str] = None,
    debin: bool = True,
    normalize: bool = False,
    **plot_params,
) -> plt.Figure:
    """
    Plot the mismatch as a function of feature.

    :param matching_data: Input data to plot.

    :param balance_calculator: Balance metric to use for calculating the per
        feature loss. Balance calculator must implement a 'per_feature_loss'
        method.

    :param reference_population: Name of population in matching_data against
        which other populations should be compared. If not supplied, will use
        the smaller population as the reference population.

    :param debin: If True, attempt to map effective features back into the real
        feature space. This is not always possible, e.g., features like
        age*height can't be mapped back to a single feature but features like
        country_US, country_Germany can. In the former case, routine will plot
        loss per effective feature; in the latter, loss per input feature.

    :param normalize: If True, divide loss by number of features such that the
        sum is the total loss. Otherwise, the plotted loss contributions must be
        averaged to obtain the total loss.

    :param plot_params: Parameters passed on to seaborn routines.
    """
    if reference_population is None:
        reference_population = _get_reference_population(matching_data)

    default_params = {
        "hue": matching_data.population_col,
        "hue_order": _get_default_hue_order(matching_data),
    }
    default_params.update(plot_params)

    # Map original features to effective features, if requested and possible
    effective_features = balance_calculator.preprocessor.output_headers["all"]
    input_output_column_mapping = dict((c, [c]) for c in effective_features)
    if debin:
        try:
            input_output_column_mapping = dict(
                (c, balance_calculator.preprocessor.get_feature_names_out(c))
                for c in matching_data.headers["all"]
            )
        except NotImplementedError:
            raise NotImplementedError(
                "Debinning not possible with given balance calculator."
            )

    indices = _debin_features(effective_features, input_output_column_mapping)

    # Compute total loss and per feature loss for all populations
    total_losses = {}
    records = []
    for population in default_params["hue_order"]:
        data = matching_data.get_population(population)
        total_losses[population] = balance_calculator.distance(data)

        per_effective_feature_loss = np.abs(
            balance_calculator.per_feature_loss(data).cpu().numpy()[0]
        )
        if normalize:
            per_effective_feature_loss /= len(
                effective_features
            )  # Normalize per feature

        # Aggregate loss per physical feature
        per_feature_loss = []
        for feature in indices.keys():
            per_feature_loss.append(per_effective_feature_loss[indices[feature]].sum())

        df = pd.DataFrame.from_dict(
            {
                "mismatch": per_feature_loss,
                "feature": list(indices.keys()),
                matching_data.population_col: [population] * len(indices.keys()),
            }
        )
        # Sort by descending mismatch wrt pool
        if population == reference_population:
            df = df.sort_values("mismatch", ascending=False)
        records.append(df)
    records = pd.concat(records)

    # Plot results
    figsize = default_params.pop("figsize", (len(indices.keys()) / 2, 6))
    fig = plt.figure(figsize=figsize)
    sns.barplot(data=records, x="feature", y="mismatch", ax=fig.gca(), **default_params)

    plt.grid(axis="y")
    handles, labels = fig.gca().get_legend_handles_labels()
    labels = [
        f"{g} ({balance_calculator.name} = {total_losses[g]:.3f})"
        for g in default_params["hue_order"]
    ]

    fig.gca().legend(handles, labels)
    plt.title(f"Contribution to {balance_calculator.name} by feature")

    plt.ylim(ymin=0)
    ymin, ymax = fig.gca().get_ylim()
    plt.vlines(plt.xticks()[0] + 0.5, ymin, ymax, linewidth=0.5, color="k")
    plt.vlines(plt.xticks()[0][0] - 0.5, ymin, ymax, linewidth=0.5, color="k")

    plt.xticks(rotation=90)
    xmin, xmax = plt.xticks()[0][0] - 0.5, plt.xticks()[0][-1] + 0.5
    plt.hlines(0.1, xmin, xmax, linewidth=2.5, color="k", linestyle="--")
    plt.xlim(xmin, xmax)

    return fig


def plot_joint_numeric_categoric_distributions(
    matching_data: MatchingData,
    include_only_numeric: Optional[List[str]] = None,
    include_only_categoric: Optional[List[str]] = None,
    **plot_params,
):
    """
    Plot 2D distributions of pairs of numeric and categoric features from
    matching_data. Choose subsets of features using include_only. Additional
    keyword arguments are passed to sns.JointGrid and override default.
    """
    default_params = {
        "hue": matching_data.population_col,
        "hue_order": _get_default_hue_order(matching_data),
    }
    default_params.update(plot_params)
    grids = []

    headers_numeric = include_only_numeric or matching_data.headers["numeric"]
    headers_categoric = include_only_categoric or matching_data.headers["categoric"]

    for x, y in itertools.product(headers_categoric, headers_numeric):
        g = sns.JointGrid(data=matching_data.data, x=x, y=y, **default_params)
        grids.append(g)

        g.plot_joint(sns.violinplot, s=25, split=True, saturation=0.9, dodge=True)
        g.ax_joint.grid(True)

        sns.histplot(
            data=matching_data.data,
            x=x,
            multiple="dodge",
            shrink=0.8,
            common_norm=False,
            discrete=True,
            stat="probability",
            ax=g.ax_marg_x,
            **default_params,
        )
        sns.histplot(
            data=matching_data.data,
            y=y,
            common_norm=False,
            stat="probability",
            ax=g.ax_marg_y,
            **default_params,
        )

        g.ax_marg_x.legend_.remove()
        g.ax_marg_y.legend_.remove()

    return grids


def plot_joint_numeric_distributions(
    matching_data: MatchingData,
    joint_kind: str = "kde",
    include_only: Optional[List[str]] = None,
    **plot_params,
):
    """
    Plot 2D distributions of pairs of numeric features from matching_data.
    joint_kind can be either kde or scatter. scatter is usually a bad choice
    for large datasets.  Choose subsets of features using include_only. Additional keyword arguments are passed to sns.JointGrid
    and override default.
    """
    default_params = {
        "hue": matching_data.population_col,
        "hue_order": _get_default_hue_order(matching_data),
    }
    default_params.update(plot_params)
    grids = []

    headers = include_only or matching_data.headers["numeric"]

    for x, y in itertools.combinations(headers, 2):
        g = sns.JointGrid(data=matching_data.data, x=x, y=y, **default_params)
        grids.append(g)

        if joint_kind == "kde":
            g.plot_joint(sns.kdeplot, levels=5)

        elif joint_kind == "scatter":
            # Give larger populations more alpha, so they don't overwhelm the plot
            counts = matching_data.counts()
            weights = (1 - counts / counts.sum()).reset_index()
            alpha = np.clip(
                matching_data.data.merge(weights, on=matching_data.population_col)[
                    "N"
                ].values,
                0.25,
                1,
            )
            g.plot_joint(sns.scatterplot, s=25, alpha=alpha)

        else:
            raise NotImplementedError(f"Unsupported joint_kind: {joint_kind}.")

        g.ax_joint.grid(True)
        g.plot_marginals(sns.histplot, bins=20, common_norm=False, stat="probability")

    return grids



from pybalance.utils import split_target_pool
from pybalance.sim import load_paper_dataset, generate_toy_dataset


def test_loader():
    m = load_paper_dataset()
    target, pool = split_target_pool(m)
    assert len(pool) == 250000
    assert len(target) == 25000


def test_generator1():
    m = generate_toy_dataset(n_pool=12345, n_target=234)
    target, pool = split_target_pool(m)
    assert len(pool) == 12345
    assert len(target) == 234



from __future__ import annotations
from typing import List, Union, Dict, Optional
import copy
from dataclasses import dataclass
import pandas as pd
import logging

logger = logging.getLogger(__name__)


@dataclass
class MatchingHeaders(object):
    """
    MatchingHeaders is a simple data structure to store information about which
    features to be used for matching and separating these features into
    categoric (e.g. country, gender) and numeric (e.g. age, weight) types.

    :param categoric: List of features to be treated as categoric variables.
    :param numeric:  List of features to be treated as numeric variables.
    """

    categoric: List[str]
    numeric: List[str]

    @property
    def all(self):
        return self.categoric + self.numeric

    # for backward compatability
    def __getitem__(self, key):
        return {"numeric": self.numeric, "categoric": self.categoric, "all": self.all}[
            key
        ]


def infer_matching_headers(
    data: pd.DataFrame,
    max_categories: int = 10,
    ignore_cols: List[str] = ["patient_id", "patientid", "population", "index_date"],
) -> MatchingHeaders:
    """
    This utility function guesses which columns are numeric and which columns
    are categoric from input data. The data can be passed either as separate
    data frames target and pool or combined in one and passed with keyword
    argument data. The function returns a dictionary with keys 'numeric',
    'categoric' and 'all' with values equal to the list of column names of the
    given type. By default, the function ignores patient_id and population
    columns.
    """
    usecols = [c for c in data.columns if c not in ignore_cols]
    data = data[usecols]

    categoric_cols = data.columns[data.nunique() <= max_categories].values.tolist()
    numeric_cols = []
    proposed_numeric = data.columns[data.nunique() > max_categories].values.tolist()
    for col in proposed_numeric:
        try:
            data[col].astype(float)
        except ValueError:
            # If column cannot be cast to numeric, treat as categoric
            logger.warning(
                f"Unable to cast {col} to float. Treating as categoric with {data[col].nunique()} categories."
            )
            categoric_cols.append(col)
        else:
            numeric_cols.append(col)

    headers = MatchingHeaders(numeric=numeric_cols, categoric=categoric_cols)

    logger.debug(f"Inferred headers: {headers}")

    return headers


def _make_quantile_function(q):
    def f(x):
        return x.quantile(q)

    if q == 0:
        name = "min"
    elif q == 1:
        name = "max"
    elif q == 0.5:
        name = "median"
    else:
        name = f"q{int(100 * q)}"
    f.__name__ = name
    return f


def _load_matching_data(path):
    if path.endswith(".csv") or path.endswith("csv.gz"):
        data = pd.read_csv(path)
    elif path.endswith(".parquet"):
        data = pd.read_parquet(path)
    else:
        raise ValueError(f"Unknown file format: {path}.")
    return data


class MatchingData(object):
    """
    It is common in matching problems to require basic metadata about the data
    in order to perform matching. For instance, the data may contain columns
    such as "patient_id", "population" and "index_date", which are not intended
    to be used for matching but which must "go along for the ride" and follow
    the main data everywhere. MatchingData is a wrapper around pandas.DataFrame
    that includes this additional required logic about the columns. Features
    required for matching are described by a "headers" field, while other
    columns exist alongside. See MatchingHeaders.

    :param data: Data frame containing both matching feature data for all
        populations as well as at least one additional column specifying to
        which population each row belongs. If a string is passed, it is assumed
        to be a path to the data frame.

    :param headers: A MatchingHeaders object with keys "numeric" and "categoric"
        and whoses values are names of columns to be used for matching. If None
        is passed, headers will be inferred based on how many unique values each
        column has. As guessing the headers can lead to errors, it is
        recommended to supply them explicitly.

    :param population_col: Name of the column used to split data into
        subpopulations.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        headers: Optional[MatchingHeaders] = None,
        population_col: str = "population",
    ):
        # load data frames if paths are passed
        if isinstance(data, str):
            data = _load_matching_data(data)
        self._data = data
        self.population_col = population_col
        self._set_headers(headers)

        if population_col not in data.columns:
            raise KeyError(
                f"""
            Cannot split into populations based on {population_col}. Column not
            present in data frame.
            """
            )

    def _set_headers(
        self, headers: Optional[Union[Dict[str, List[str]], MatchingHeaders]]
    ) -> None:
        """
        Set private data needed to construct the headers property. Headers are
        set at initialization and are considered immutable. If you need to
        change the headers, you must create a new instance of MatchingData.
        """
        if headers is None:
            headers = infer_matching_headers(
                self.data,
                ignore_cols=[
                    "patient_id",
                    "patientid",
                    "index_date",
                    self.population_col,
                ],
            )
        elif not isinstance(headers, MatchingHeaders):
            headers = MatchingHeaders(**headers)

        self.headers = headers

    def get_population(self, population: str) -> pd.DataFrame:
        """
        Get the matching data for a population by its name.
        """
        pop = self._data[self._data[self.population_col] == population]
        if not len(pop):
            raise KeyError(f"Population {population} not found!")
        return pop

    @property
    def populations(self) -> List[str]:
        """
        List of all populations present in the MatchingData object.
        """
        return sorted(self[self.population_col].unique().tolist())

    @property
    def data(self) -> pd.DataFrame:
        """
        Pointer to underlying pandas DataFrame.
        """
        # Defining data as a property prevents the user from manually setting
        # the data.
        return self._data

    def sample(self, n: int = 5) -> pd.DataFrame:
        """
        Sample underlying pandas DataFrame.
        """
        return self.data.sample(n=n)

    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return first n rows from underlying pandas DataFrame.
        """
        return self.data.head(n=n)

    def tail(self, n: int = 5) -> pd.DataFrame:
        """
        Return last n rows from underlying pandas DataFrame.
        """
        return self.data.tail(n=n)

    def __getitem__(self, key: Union[List[str], str]):
        return self.data.__getitem__(key)

    def copy(self) -> MatchingData:
        """
        Create a new MatchingData instance with exact same data and metadata.
        """
        data = copy.deepcopy(self.data)
        headers = copy.deepcopy(self.headers)
        matching_data = MatchingData(
            data=data, headers=headers, population_col=self.population_col
        )
        return matching_data

    def append(self, df: pd.DataFrame, name: Optional[str] = None) -> None:
        """
        Append a population to an existing MatchingData instance. This operation
        is inplace.
        """
        if not set([self.population_col] + self.headers["all"]) <= set(df.columns):
            missing_columns = set([self.population_col] + self.headers["all"]) - set(
                df.columns
            )
            raise ValueError(
                f"Required columns {list(missing_columns)} are missing from data to be appended."
            )

        # copy input data to avoid undesired side effects
        df = copy.deepcopy(df)
        if name is not None:
            df.loc[:, self.population_col] = name

        self._data = pd.concat([self._data, df])

    def to_csv(self, *args, **kwargs):
        """
        Write underlying pandas DataFrame to csv. Call signature is identical to
        pandas method.
        """
        return self.data.to_csv(*args, **kwargs)

    def to_parquet(self, *args, **kwargs):
        """
        Write underlying pandas DataFrame to parquet. Call signature is
        identical to pandas method.
        """
        return self.data.to_parquet(*args, **kwargs)

    def __str__(self):
        return f"""
Headers Numeric:
{self.headers['numeric']}

Headers Categoric:
{self.headers['categoric']}

Populations:
{self.populations}

{self.data}"""

    def _repr_html_headers_(self):
        return f"""
        <b>Headers Numeric: </b><br>
        {self.headers['numeric']}<br><br>
        <b>Headers Categoric: </b><br>
        {self.headers['categoric']} <br><br>
        <b>Populations</b> <br>
        {self.populations} <br>
        """

    def _repr_html_(self):
        return self._repr_html_headers_() + self.data._repr_html_()

    def __len__(self):
        return len(self.data)

    def counts(self):
        counts = self.data.reset_index().groupby(self.population_col).count()[["index"]]
        counts.columns = ["N"]
        return counts

    def describe_numeric(
        self,
        aggregations=["mean", "std"],
        quantiles=[0, 0.25, 0.5, 0.75, 1],
        long_format=True,
    ) -> pd.DataFrame:
        """
        Create a summary statistics table split by population for numeric variables.
        """
        # numeric
        aggregations = aggregations + [_make_quantile_function(q) for q in quantiles]
        agg = dict((c, aggregations) for c in self.headers["numeric"])
        agg = self.data.reset_index().groupby(self.population_col).agg(agg).T
        agg.columns = [c for c in agg.columns]
        agg = agg.round(decimals=2)

        if not long_format:
            agg = agg.unstack(level=-1)

        return agg

    def describe_categoric(self, normalize=True) -> pd.DataFrame:
        """
        Create a summary statistics table split by population for categoric variables.
        """
        counts = self.counts()["N"]
        counts = pd.DataFrame.from_records(
            [counts.values.astype(int).tolist()],
            index=pd.MultiIndex.from_tuples([(f"{self.population_col} size", "N")]),
            columns=counts.index.values.tolist(),
        )

        # categoric
        out = [counts]
        for cat in self.headers["categoric"]:
            tmp = (
                self.data.reset_index()
                .groupby(["population", cat])
                .count()[["index"]]
                .reset_index()
            )
            tmp.loc[:, "feature"] = cat
            tmp = tmp.pivot(
                index=["feature", cat], columns=["population"], values=["index"]
            )
            tmp.columns = [c[1] for c in tmp.columns]
            tmp.index.names = ["feature", "value"]
            out.append(tmp)

        out = pd.concat(out).fillna(0)

        if normalize:
            for c in counts.columns:
                out.loc[:, c] = out[c] / counts.iloc[0][c]
        else:
            out = out.astype(int)

        return out

    def describe(
        self,
        normalize: bool = True,
        aggregations: List[str] = ["mean", "std"],
        quantiles: List[float] = [0, 0.25, 0.5, 0.75, 1],
    ) -> pd.DataFrame:
        """
        Calls describe_categoric() and describe_numeric() and returns the
        results in a single dataframe.
        """
        c = self.describe_categoric(normalize)
        n = self.describe_numeric(aggregations, quantiles)
        return pd.concat([c, n])


def split_target_pool(
    matching_data: MatchingData,
    pool_name: Optional[str] = None,
    target_name: Optional[str] = None,
) -> pd.DataFrame:
    """
    Split matching_data into target and pool populations based. If
    the names of the target and pool populations are not
    explicitly provided, the routine will attempt to infer their names,
    assuming that the target population is the smaller population.
    """

    if isinstance(target_name, str) and isinstance(pool_name, str):
        target = matching_data.get_population(target_name)
        pool = matching_data.get_population(pool_name)
    elif isinstance(target_name, str) or isinstance(pool_name, str):
        if len(matching_data.populations) != 2:
            raise ValueError(
                f"""
            Cannot split into exactly two populations based on {matching_data.population_col}.
            Found populations: {','.join(matching_data.populations)}.
            """
            )
        if isinstance(target_name, str):
            pool_name = [p for p in matching_data.populations if p != target_name][0]
        if isinstance(pool_name, str):
            target_name = [p for p in matching_data.populations if p != pool_name][0]
        target = matching_data.get_population(target_name)
        pool = matching_data.get_population(pool_name)
    else:
        if len(matching_data.populations) != 2:
            raise ValueError(
                f"""
            Cannot split into exactly two populations based on {matching_data.population_col}.
            Found populations: {','.join(matching_data.populations)}.
            """
            )
        inferred_pool_name = matching_data.populations[0]
        inferred_target_name = matching_data.populations[1]
        target = matching_data.get_population(inferred_target_name)
        pool = matching_data.get_population(inferred_pool_name)

        # bigger population considered pool, just a convention, no real effect
        if len(pool) < len(target):
            _pool = pool
            pool = target
            target = _pool

    return target, pool



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import seaborn as sns

"""
Plotting functions for history of genetic algorithm
History is a dictionary containing parallel lists
Index in lists corresponds to generation
Three current lists are beta, gamma, and candidate_populations
"""


def get_n_colors(n, cspace="viridis"):
    cmap = cm.get_cmap(cspace)
    colors = [cmap(i / n) for i in range(n)]
    colors.reverse()
    return colors


def plot_history_of_beta_and_gamma(history, metrics):
    fig, axes = plt.subplots(1, len(metrics), figsize=[10 * len(metrics), 4])
    if len(metrics) == 1:
        axes = [axes]

    for i, metric in enumerate(metrics):
        plot_density_history_of_metric(history, metric, 10, axes[i])

    return fig


def plot_density_history_of_metric(history, metric, n_steps, ax=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=[10, 4])
    n_generations = len(history[f"{metric}_balance"])
    colors = get_n_colors(n_generations)
    if n_steps > n_generations:
        n_steps = n_generations
    for g in range(0, n_generations, int(n_generations / n_steps)):
        sns.kdeplot(
            -1 * history[f"{metric}_balance"][g],
            ax=ax,
            color=colors[g],
            label=f"Gen {g}",
        )
    sns.kdeplot(
        -1 * history[f"{metric}_balance"][n_generations - 1],
        ax=ax,
        color=colors[g],
        label=f"Gen {n_generations-1}",
    )
    ax.set_xlabel(f"{metric}")
    ax.set_title(f"History of {metric} scores over generations")
    ax.legend()
    ax.grid("on")


def plot_convergence(*histories, labels=None, ax=None, metric="gamma"):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=[12, 6])

    colors = get_n_colors(len(histories))
    label = None
    for exp, history in enumerate(histories):
        best_candidate = -np.array(histories[exp][f"{metric}_balance"]).max(axis=1)
        if labels:
            label = labels[exp]

        ax.plot(best_candidate, label=label, c=colors[exp])
    ax.set_xlabel("generation")
    ax.set_ylabel(f"best {metric}")
    ax.set_title(f"Best {metric} score over generations")
    ax.grid()
    fig.legend()
    return fig



from pybalance.utils import *
import pybalance.genetic
import pybalance.propensity
import pybalance.sim
import pybalance.visualization
import pybalance.lp

__version__ = "0.1.1"

import logging

logger = logging.getLogger(__name__)
logger.info(f"Loaded pybalance version {__version__}.")

# Logging is configured at the application level. To adjust logging level,
# configure logging as below before importing pybalance:
#
# import logging
# logging.basicConfig(
#     format="%(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
#     level='INFO',
# )
#



import numpy as np

import logging

logger = logging.getLogger(__name__)

from pybalance.propensity import (
    PropensityScoreMatcher,
    propensity_score_match,
)
from pybalance.utils import split_target_pool


class GeneticMatcherInitializer:
    def __init__(self, matcher):
        self.matcher = matcher
        self.verbose = matcher.params["verbose"]
        self.params = matcher.params["initialization"]
        self.matcher_params = matcher.params
        self.balance_calculator = matcher.balance_calculator
        self._propensity = None

    def initialize(self, n_candidate_populations, seed=None):
        """
        Initialize genetic algorithm. In the genetic algorithm, we take a large
        number of candidate patient groups and evolve them under selection
        pressure from a balance function. The initial population is constructed
        here.
        """
        if self.verbose:
            logging.info("Optimizing balance with genetic algorithm ...")
            logging.info("Initial balance scores:")
            candidate_population = split_target_pool(
                self.balance_calculator.matching_data
            )[1]
            benchmark = self.balance_calculator.distance(candidate_population)
            logging.info(f"\t{self.balance_calculator.name}:\t{benchmark:3.3f}")
            logging.info("Initializing candidate populations ...")

        # initialize candidate populations
        if not seed:
            candidate_populations = []
        else:
            candidate_populations = seed

        # computing reference benchmarks and add to initial populations if requested by user
        benchmarks = self.compute_benchmark_populations()
        candidate_populations.extend(benchmarks)

        # randomly sample remaining candidate populations
        n_remaining = n_candidate_populations - len(candidate_populations)
        n_samples = self.split_samples_among_methods(
            self.params["sampling"], n_remaining
        )
        for sampling_method, N in n_samples.items():
            candidate_populations.extend(self.sample_patients(N, sampling_method))

        return candidate_populations

    def compute_benchmark_populations(self):
        benchmarks = []

        def choose_to_include(candidate_population, mode):
            if self.verbose:
                benchmark = self.balance_calculator.distance(candidate_population)
                logging.info(f"\t{self.balance_calculator.name}:\t{benchmark:3.3f}")

            if mode == "include":
                benchmarks.append(list(candidate_population))
                if self.verbose:
                    logging.info("\tIncluded in initial population.\n")
            elif self.verbose:
                logging.info("\tExcluded from initial population.\n")

            return

        for method, mode in self.params["benchmarks"].items():
            # possible modes are exclude, include or benchmark
            # exclude -- do no nothing
            # inlucde -- compute benchmark, add to initial population
            # benchmark -- compute benchmark, do not add to initial population
            if mode == "exclude":
                continue

            if self.verbose:
                logging.info(f"Computing {method.upper()} 1-1 matching method ...")

            if method == "propensity":
                # scipy appears to be modifying input arrays inplace - thus one MUST copy the arrays
                # here, since we are not done with propensity scores yet
                propensity_pool, propensity_target = (
                    self.propensity["pool"],
                    self.propensity["target"],
                )
                candidate_population, _ = propensity_score_match(
                    np.copy(propensity_pool),
                    np.copy(propensity_target),
                    method="greedy",
                )
                choose_to_include(candidate_population, mode)

            else:
                raise (
                    NotImplementedError(
                        f"Unknown initialization method method {method}."
                    )
                )

        return benchmarks

    def sample_patients(self, N, method="uniform"):
        candidate_populations = []
        if not N:
            # some of the downstream calculations (e.g. mahalanobis matrix) are
            # mildly expensive and so we should short circuit here if we aren't
            # actually needing any samples
            return candidate_populations

        # Get sampling probabilities for the patients from the pool
        idx = list(range(len(self.balance_calculator.pool)))
        if method == "uniform":
            init_probs = np.array([1] * len(idx))
        elif method == "propensity":
            init_probs = self.propensity["pool"]

        # this is a hacky form of regularization, don't allow the model to be
        # over confident about excluding patients
        init_probs = np.clip(init_probs, 1e-8, 1)
        p = init_probs / init_probs.sum()

        if self.verbose:
            logging.info(
                f"Sampling {N} candidate populations according to {method.upper()} distribution ...\n"
            )

        for _ in range(N):
            candidate_population = np.random.choice(
                idx,
                size=self.matcher_params["candidate_population_size"],
                p=p,
                replace=False,
            )
            candidate_populations.append(candidate_population)

        # p = torch.FloatTensor(p)
        # weights = p.repeat(N, 1)
        # candidate_populations = torch.multinomial(weights, num_samples=self.matcher_params['candidate_population_size']).cpu().numpy()

        return candidate_populations

    def split_samples_among_methods(self, sampling_methods, N):
        """
        Spread N samples among the chosen sampling_methods to approximately the
        correct proportions.
        """
        total = sum(sampling_methods.values())
        target = N
        for method in list(sampling_methods.keys())[:-1]:
            sampling_methods[method] = round(sampling_methods[method] * N / total)
            target -= sampling_methods[method]
        method = list(sampling_methods.keys())[-1]
        sampling_methods[method] = target
        return sampling_methods

    @property
    def propensity(self):
        if self._propensity is None:
            ps_matcher = PropensityScoreMatcher(
                matching_data=self.balance_calculator.matching_data,
                objective=self.balance_calculator,
            )
            ps_matcher.match()
            propensity_pool, propensity_target = ps_matcher.get_propensity_score()
            self._propensity = {"pool": propensity_pool, "target": propensity_target}
        return self._propensity



import pandas as pd
import numpy as np

from pybalance.utils.preprocess import (
    FloatEncoder,
    CategoricOneHotEncoder,
    NumericBinsEncoder,
    DecisionTreeEncoder,
    ChainPreprocessor,
)
from pybalance.sim import generate_toy_dataset
from pybalance.utils import MatchingData, MatchingHeaders


def test_float_encoder():
    N = 100
    np.random.seed(1234)
    data = pd.DataFrame.from_dict(
        {
            "cat1": np.random.choice(list("abcdefghijk"), size=N, replace=True),
            "cat1": np.random.choice([1, 2, 3, 4, 5, 10], size=N, replace=True),
        }
    )
    data.loc[:, "population"] = ["pool"] * len(data)
    data.loc[:, "patient_id"] = list(range(len(data)))
    matching_data = MatchingData(data=data)

    pp_float = FloatEncoder()
    transformed_data = pp_float.fit_transform(matching_data)

    assert (
        transformed_data.headers["categoric"] == transformed_data.headers["categoric"]
    )
    assert transformed_data.headers["numeric"] == transformed_data.headers["numeric"]

    # test that the transformed data has same number of unique values as the
    # original data
    assert data["cat1"].nunique() == data["cat1"].nunique()

    # test that the transformed values align with old values
    combined_data = data.merge(
        transformed_data.data, on="patient_id", suffixes=["_old", "_new"]
    )
    for val, df in combined_data.groupby("cat1_old"):
        assert df["cat1_new"].nunique() == 1


def test_categoric_encoder():
    data = pd.DataFrame.from_dict(
        {"cat1": ["a", "b", "c", "b"], "cat2": [1, 2, 3, 4], "bin1": [0, 1, 0, 1]}
    )
    data.loc[:, "population"] = ["pool"] * len(data)
    data.loc[:, "patient_id"] = list(range(len(data)))
    matching_data = MatchingData(data=data)

    pp_cat = CategoricOneHotEncoder(drop="if_binary")
    transformed_data = pp_cat.fit_transform(matching_data)
    assert len(transformed_data.headers["numeric"]) == len(
        matching_data.headers["numeric"]
    )

    # check correct number of output columns
    n_unique_categories = matching_data[matching_data.headers["categoric"]].nunique()
    n_columns_expected = n_unique_categories[n_unique_categories > 2].sum() + len(
        n_unique_categories[n_unique_categories == 2]
    )
    assert len(transformed_data.headers["categoric"]) == n_columns_expected

    # check all output columns are now binary
    assert all(transformed_data[transformed_data.headers["categoric"]].nunique() == 2)

    # Check actual output values
    # Technically not really a great test, since the column naming might not be stable.
    for cat in matching_data.headers["categoric"]:
        if matching_data[cat].nunique() > 2:
            for val in matching_data[cat].unique():
                assert all(
                    transformed_data[f"{cat}_{val}"] == (data[cat] == val).astype(int)
                )

    # check transforming on data with missing categories still includes missing
    # categories in the output
    data_with_missing_categories = pd.DataFrame.from_dict(
        {"cat1": ["a", "b", "b"], "cat2": [1, 2, 4], "bin1": [0, 1, 1]}
    )
    data_with_missing_categories.loc[:, "population"] = ["pool"] * len(
        data_with_missing_categories
    )
    data_with_missing_categories.loc[:, "patient_id"] = list(
        range(len(data_with_missing_categories))
    )
    matching_data = MatchingData(data=data_with_missing_categories)
    transformed_data = pp_cat.transform(matching_data)
    assert len(transformed_data.headers["categoric"]) == n_columns_expected


def test_categoric_noop():
    # Test when there are no categoric variables that the preprocessor does
    # nothing

    data1 = pd.DataFrame.from_dict({"col": np.linspace(0, 1, 2001)})
    data2 = pd.DataFrame.from_dict({"col": np.linspace(0.5, 1.5, 1001)})
    data1.loc[:, "population"] = "pool"
    data2.loc[:, "population"] = "target"

    matching_data = MatchingData(pd.concat([data1, data2]))
    pp = CategoricOneHotEncoder()
    m_out = pp.fit_transform(matching_data)
    assert np.all(
        m_out[matching_data.headers.all] == matching_data[matching_data.headers.all]
    )


def test_categoric_drop():
    data = pd.DataFrame.from_dict(
        {
            "population": ["a", "b", "a", "b"],
            "gender": ["Male", "Female", "Male", "Male"],
            "country": ["1", "2", "3", "4"],
        }
    )
    m = MatchingData(data)
    pp = CategoricOneHotEncoder(drop="first")
    pp.fit_transform(m)
    assert pp.get_feature_names_out("gender") == ["gender_Male"]
    assert pp.get_feature_names_out("country") == [
        "country_2",
        "country_3",
        "country_4",
    ]


def test_numeric_bins_encoder():
    N = 10
    df = pd.DataFrame.from_dict({"x": list(range(N)), "population": ["x"] * N})
    headers = MatchingHeaders(numeric=["x"], categoric=[])
    m = MatchingData(df, headers=headers)
    p = NumericBinsEncoder(n_bins=N)
    df_transformed = p.fit_transform(m)
    assert df_transformed["x"].nunique() == N
    for i in range(N):
        assert df_transformed.data.loc[i]["x"] == i


def test_numeric_bins_encoder_feature_names():
    # Test number of feature names out returned == number of bins. Since the
    # tested method is based on a regexp, try to use some confusing column
    # names.

    # Start from basic dataset
    m = generate_toy_dataset()
    data = m.data

    # Add features that might confuse the regexp
    data.loc[
        :, "age"
    ] = 1  # This will make the number of returned bins < number requested
    data.loc[:, "age_weight"] = data["age"] * data["weight"]
    data.loc[:, "age*weight"] = (
        data["age"] * data["weight"]
    )  # include special regexp symbol
    data.loc[:, "age-weight"] = data["age"] * data["weight"]
    m = MatchingData(data)

    N = 15
    pp_numeric = NumericBinsEncoder(n_bins=N, cumulative=False, encode="onehot")
    pp_numeric.fit_transform(m)

    for feature_name_in in m.headers.numeric:
        mapped_cols = pp_numeric.get_feature_names_out(feature_name_in)
        feature_index = pp_numeric.discretizer.feature_names_in_.tolist().index(
            feature_name_in
        )
        assert len(mapped_cols) == pp_numeric.discretizer.n_bins_[feature_index]


def test_numeric_bins_encoder_feature_names_order():
    # Test that the order of the feature names out imply increasing values. This
    # is a prerequisite for generating the cumulative bins.

    # Start from basic dataset
    m = generate_toy_dataset()
    data = m.data

    # add a feature that might confuse the order of bins
    data.loc[:, "age"] = -data["age"]
    data.loc[:, "age*weight"] = data["age"] * data["weight"]
    m = MatchingData(data)

    N = 15
    pp_numeric = NumericBinsEncoder(n_bins=N, cumulative=False, encode="onehot")
    m_out = pp_numeric.fit_transform(m)

    for feature_name_in in m.headers.numeric:
        mapped_cols = pp_numeric.get_feature_names_out(feature_name_in)
        for left, right in zip(mapped_cols[:-1], mapped_cols[1:]):
            left_max = m.data.loc[m_out[left] == 1, feature_name_in].max()
            right_min = m.data.loc[m_out[right] == 1, feature_name_in].min()
            assert left_max < right_min


def test_numeric_bins_encoder_onehot_cumulative():
    # Test the cumulative bins are generated correctly.

    # Start from basic dataset
    m = generate_toy_dataset()
    data = m.data

    # add a feature that might confuse the order of bins
    data.loc[:, "age"] = -data["age"]
    data.loc[:, "age*weight"] = data["age"] * data["weight"]
    m = MatchingData(data)

    N = 15
    pp_numeric = NumericBinsEncoder(n_bins=N, cumulative=True, encode="onehot")
    m_out = pp_numeric.fit_transform(m)

    # test that the ORDER of the feature names out imply increasing values
    for feature_name_in in m.headers.numeric:
        mapped_cols = pp_numeric.get_feature_names_out(feature_name_in)
        for left, right in zip(mapped_cols[:-1], mapped_cols[1:]):
            # test that left == 1 --> right == 1
            right_min = m_out.data.loc[m_out[left] == 1, right].min()
            assert right_min == 1


def test_numeric_bins_encoder_onehot_cumulative2():
    data = pd.DataFrame.from_dict(
        {"age": [1, 2, 3, 4, 5, 6.0, 7, 8, 9, 10, 11, 2, 3, 4]}
    )
    data.loc[:, "population"] = ["a"] * len(data["age"])
    m = MatchingData(data, MatchingHeaders(numeric=["age"], categoric=[]))

    pp = NumericBinsEncoder(cumulative=True, n_bins=10, strategy="quantile")
    pp.fit(m)
    m_out = pp.transform(m)

    cols_out = pp.get_feature_names_out("age")
    # check last column is dropped
    assert m_out.data.loc[m_out["age"] == 10, cols_out].values.max() == 0
    assert m_out.data.loc[m_out["age"] == 1, cols_out].values.min() == 1


def test_decision_tree_encoder():
    m = generate_toy_dataset()
    pp_float = DecisionTreeEncoder(max_leaf_nodes=3)
    m_out = pp_float.fit_transform(m)
    assert len(m_out.headers.all) == 3

    pp_float = DecisionTreeEncoder(keep_original_features=True, max_leaf_nodes=3)
    m_out = pp_float.fit_transform(m)
    assert len(m_out.headers.all) == 3 + len(m.headers.all)


def test_chain_preprocessor_output_names():
    m = generate_toy_dataset()
    pp_n = NumericBinsEncoder(encode="onehot", n_bins=6)
    pp_c = CategoricOneHotEncoder(drop=None)
    pp = ChainPreprocessor([pp_c, pp_n])
    pp.fit(m)

    expected = {
        "gender": ["gender_0.0", "gender_1.0"],
        "binary_0": ["binary_0_0", "binary_0_1"],
        "binary_1": ["binary_1_0", "binary_1_1"],
        "binary_2": ["binary_2_0", "binary_2_1"],
        "binary_3": ["binary_3_0", "binary_3_1"],
        "country": [
            "country_0",
            "country_1",
            "country_2",
            "country_3",
            "country_4",
            "country_5",
        ],
        "haircolor": ["haircolor_0", "haircolor_1", "haircolor_2"],
        "age": ["age_0.0", "age_1.0", "age_2.0", "age_3.0", "age_4.0", "age_5.0"],
        "weight": [
            "weight_0.0",
            "weight_1.0",
            "weight_2.0",
            "weight_3.0",
            "weight_4.0",
            "weight_5.0",
        ],
        "height": [
            "height_0.0",
            "height_1.0",
            "height_2.0",
            "height_3.0",
            "height_4.0",
            "height_5.0",
        ],
    }

    for c in m.headers["all"]:
        assert pp.get_feature_names_out(c) == expected[c]


def test_feature_names_out():
    pp = NumericBinsEncoder(cumulative=True, n_bins=5)
    m = generate_toy_dataset()
    pp.fit_transform(m)
    assert set(pp.get_feature_names_out()) == set(pp.output_headers["all"])



from .matching_data import (
    MatchingData,
    MatchingHeaders,
    infer_matching_headers,
    split_target_pool,
)
from .preprocess import (
    BaseMatchingPreprocessor,
    ChainPreprocessor,
    FloatEncoder,
    NumericBinsEncoder,
    CategoricOneHotEncoder,
    DecisionTreeEncoder,
    StandardMatchingPreprocessor,
    GammaPreprocessor,
    CrossTermsPreprocessor,
    GammaXPreprocessor,
    BetaXPreprocessor,
)
from .balance_calculators import (
    BalanceCalculator,
    BaseBalanceCalculator,
    BatchedBalanceCaclulator,
    BetaBalance,
    BetaSquaredBalance,
    BetaMaxBalance,
    GammaBalance,
    GammaSquaredBalance,
    GammaXTreeBalance,
    GammaXBalance,
    BetaXBalance,
    map_input_output_weights,
    BALANCE_CALCULATORS,
)
from .misc import require_fitted



import numpy as np
import time
import torch

import logging

logger = logging.getLogger(__name__)


class BasicLogger:
    """
    BasicLogger class for use with GeneticMatcher class. All loggers must
    implement at least two methods: on_generation_end, which is called at the
    end of each generation, and on_matching_end, which is called at the end of
    all iterations.  These methods must accept a GeneticMatcher instance as the
    only argument.
    """

    def __init__(self, log_every=10, working_directory="results/"):
        self.log_every = log_every
        self._t0 = time.time()
        self._set_directories(working_directory)

    def _set_directories(self, working_directory):
        self.working_directory = working_directory
        # self.history_directory = os.path.join(self.working_directory, 'history')
        # pathlib.Path(self.history_directory).mkdir(parents=True, exist_ok=True)

    def on_generation_end(self, matcher):
        verbose = matcher.params["verbose"]

        unique_patients = torch.unique(matcher.candidate_populations)
        if verbose and not (matcher.generation % self.log_every):
            logger.info(f"Generation {matcher.generation}")
            logger.info(f"\tremaining patients: {len(unique_patients)}")
            logger.info(f"\telapsed time: {(time.time() - self._t0)/60:.2f} min")

        balance_tensor = matcher.balance.cpu().numpy()
        idx_best_match = balance_tensor.argmax()
        best_score = np.abs(balance_tensor[idx_best_match])
        idx_worst_match = balance_tensor.argmin()
        worst_score = np.abs(balance_tensor[idx_worst_match])

        if verbose and not (matcher.generation % self.log_every):
            logger.info(
                f"\tbest {matcher.balance_calculator.name}: {best_score:3.5f} \tworst {matcher.balance_calculator.name}: {worst_score:3.5f}"
            )

    def on_matching_end(self, matcher):
        # self.make_figures(matcher)
        self.save_results_locally(matcher)

    def make_figures(self, matcher):
        # # initial distributions
        # fig = plot_1d_marginals(matcher.matching_data)
        # figname = os.path.join(self.working_directory, 'fig_initial_distributions.png')
        # fig.savefig(figname)

        # # final distributions
        # best_match = matcher.get_best_match()
        # matching_data = matcher.matching_data.append(best_match)
        # fig = plot_1d_marginals(best_match, matcher.target, matcher.headers)
        # figname = os.path.join(self.working_directory, 'fig_final_distributions.png')
        # fig.savefig(figname)

        # # metrics history / convergence
        # figname = os.path.join(self.working_directory, 'fig_distribution_of_gamma_and_beta.png')
        # # fig = plot_history_of_beta_and_gamma(matcher.history, matcher.params['balance_metrics'])
        # fig.savefig(figname)
        pass

    def save_results_locally(self, matcher):
        pass

    def save_results_remotely(self, matcher):
        pass



import pandas as pd
import numpy as np
import copy
import pytest

from pybalance.sim import generate_toy_dataset
from pybalance.utils import (
    MatchingData,
    MatchingHeaders,
    split_target_pool,
    infer_matching_headers,
)


def _make_fake_matching_data(
    N_pool: int, N_target: int, pool_name: str = "pool", target_name="target"
) -> pd.DataFrame:
    # data shape
    N_numeric_features = 20
    N_categoric_features = 8

    # numeric data
    pool_n = np.random.rand(N_pool, N_numeric_features)
    target_n = np.random.rand(N_target, N_numeric_features)

    # categoric data
    pool_c = np.random.choice([1, 2, 3], size=(N_pool, N_categoric_features - 2))
    target_c = np.random.choice([1, 2, 3], size=(N_target, N_categoric_features - 2))

    # these have a lot of categories, but should be identified as categoric
    # since they cannot be cast to float
    pool_c2 = np.random.choice(list("abcdefghijklmnopqrstuvwxyz"), size=(N_pool, 2))
    target_c2 = np.random.choice(list("abcdefghijklmnopqrstuvwxyz"), size=(N_target, 2))

    # column names
    numeric_cols = [f"num_col_{d}" for d in range(N_numeric_features)]
    categoric_cols = [f"cat_col_{d}" for d in range(N_categoric_features)]

    # all data
    pool = np.hstack((pool_n, pool_c, pool_c2))
    target = np.hstack((target_n, target_c, target_c2))
    pool = pd.DataFrame.from_records(pool, columns=numeric_cols + categoric_cols)
    target = pd.DataFrame.from_records(target, columns=numeric_cols + categoric_cols)
    pool.loc[:, "population"] = pool_name
    target.loc[:, "population"] = target_name
    data = pd.concat([pool, target])
    data.loc[:, "patient_id"] = list(range(len(data)))

    return data


def test_split_target_pool():
    m = generate_toy_dataset()

    # test we can correctly infer the populations based on size
    target, pool = split_target_pool(m)
    assert len(target) == 1000
    assert len(pool) == 10000
    assert target[m.population_col].unique() == ["target"]
    assert pool[m.population_col].unique() == ["pool"]

    # test we can override the guessing behavior by explicitly giving the
    # poplation names
    target, pool = split_target_pool(m, target_name="pool", pool_name="target")
    assert len(target) == 10000
    assert len(pool) == 1000

    # test we can find the populations when extra ones are in there
    target, pool = split_target_pool(m, target_name="pool", pool_name="target")
    assert len(target) == 10000
    assert len(pool) == 1000


def test_split_target_pool2():
    N_pool, N_target = 1000, 100
    pool_name, target_name = "p", "t"
    data = _make_fake_matching_data(N_pool, N_target, pool_name, target_name)

    matching_data = MatchingData(data, population_col="population")
    target, pool = split_target_pool(matching_data)

    assert len(target) == N_target
    assert len(pool) == N_pool
    assert pool_name == pool["population"].unique()[0]
    assert target_name == target["population"].unique()[0]
    assert len(matching_data.data) == len(data)


def test_split_target_pool3():
    N_pool, N_target = 500, 250
    pool_name, target_name = "p", "t"
    data = _make_fake_matching_data(N_pool, N_target, pool_name, target_name)

    matching_data = MatchingData(data, population_col="population")

    # test we can find the right populations when more than two are present
    match = matching_data.get_population(pool_name).sample(N_target)
    matching_data.append(match, name="match")
    target, pool = split_target_pool(
        matching_data, target_name=target_name, pool_name=pool_name
    )
    assert target_name == target["population"].unique()[0]
    assert pool_name == pool["population"].unique()[0]


def test_split_target_pool4():
    # Test can split when two populations present and only one name passed
    m = generate_toy_dataset()
    target, pool = split_target_pool(m, target_name="pool")
    assert target[m.population_col].unique()[0] == "pool"


def test_get_population():
    m = generate_toy_dataset()

    # Check we can lookup populations by name
    pool = m.get_population("pool")
    assert len(pool) == 10000
    assert pool[m.population_col].unique() == ["pool"]

    # check that a KeyError is raised when population does not exists
    with pytest.raises(KeyError):
        m.get_population("abcdefg")


def test_describe():
    m = generate_toy_dataset()
    quantiles = [0, 0.25, 0.5, 0.75, 1]
    summary = m.describe(aggregations=[], quantiles=quantiles)
    target, pool = split_target_pool(m)
    for c in m.headers["numeric"]:
        assert np.allclose(
            summary["pool"][c].values, pool[c].quantile(quantiles), atol=1e-2
        )
        assert np.allclose(
            summary["target"][c].values, target[c].quantile(quantiles), atol=1e-2
        )


def test_infer_matching_headers():
    size = 1000
    np.random.seed(1234)
    data = pd.DataFrame.from_dict(
        {
            "patient_id": list(range(size)),
            "cat1": np.random.choice(list("abcdef"), size=size, replace=True),
            "cat2": np.random.choice(list(range(6)), size=size, replace=True),
            "cat3": np.random.choice(list(range(3)), size=size, replace=True),
            "num1": np.random.rand(size),
            "num2": np.random.randn(size),
        }
    )

    # cat1, cat2, cat3 should be inferred to be categoric and the rest are
    # numeric; patient_id is a special column that is ignored
    headers = infer_matching_headers(data, max_categories=10)
    assert set(headers.categoric) == set(["cat1", "cat2", "cat3"])
    assert set(headers.numeric) == set(["num1", "num2"])

    # cat2 should now be considered numeric, since it has too many categories
    # cat1 has too many categories, but can't be cast to numeric so it's still
    # categoric
    headers = infer_matching_headers(data, max_categories=3)
    print(headers)
    assert set(headers.categoric) == set(["cat1", "cat3"])
    assert set(headers.numeric) == set(["cat2", "num1", "num2"])


def test_matching_data_append():
    N_pool, N_target = 1000, 100
    pool_name, target_name = "p", "t"
    data = _make_fake_matching_data(N_pool, N_target, pool_name, target_name)
    matching_data = MatchingData(data, population_col="population")

    # test append
    match = matching_data.get_population(pool_name).sample(N_target)
    matching_data.append(match, name="match")
    assert set(matching_data.populations) == set(["match", pool_name, target_name])


def test_matching_data_append2():
    N_pool, N_target = 1000, 100
    pool_name, target_name = "p", "t"
    data = _make_fake_matching_data(N_pool, N_target, pool_name, target_name)

    # test append fails with missing columns
    new_data = copy.deepcopy(data)
    new_data.loc[:, "adfadsf"] = 1
    matching_data = MatchingData(new_data)
    with pytest.raises(ValueError):
        matching_data.append(data)


def test_matching_data_copy():
    # check there are no side effects when making changes on a copy

    m1 = generate_toy_dataset()
    m2 = m1.copy()

    pool1 = m1.get_population("pool")
    orig_value = pool1["age"][0]
    pool2 = m2.get_population("pool")
    pool2.loc[:, "age"] = -100

    new_value = pool1["age"][0]
    assert orig_value == new_value
    assert pool2["age"][0] == -100


def test_specified_headers():
    N = 100
    df = pd.DataFrame.from_dict(
        {
            "cat3": np.random.choice(list(range(3)), size=N, replace=True),
            "population": ["x"] * N,
        }
    )

    # Force cat3 to be considered numeric, even though it would be inferred as
    # categoric
    headers = MatchingHeaders(numeric=["cat3"], categoric=[])
    m = MatchingData(df, headers=headers)
    assert m.headers["categoric"] == []
    assert m.headers["numeric"] == ["cat3"]



import numpy as np


from pybalance.lp import ConstraintSatisfactionMatcher
from pybalance.sim import generate_toy_dataset


def sum_feature(pool, patients, f):
    patients = [pool[p] for p in patients]
    return sum(p[f] for p in patients)


def test_limits():
    # basically test that the limits chosen by the solver are feasible
    # in the ideal case, ALL subsets should satisfy the constraint.
    matching_data = generate_toy_dataset()

    solver = ConstraintSatisfactionMatcher(matching_data, objective="beta")

    min_features = []
    max_features = []
    max_deltas = []
    match_size = int(len(solver.target) / 2)
    n_features = len(solver.pool[0])

    for f in range(n_features):
        min_feature, max_feature, max_delta = solver._get_limits(
            f, match_size=match_size
        )
        min_features.append(min_feature)
        max_features.append(max_feature)
        max_deltas.append(max_delta)

    patient_pool_list = list(range(len(solver.pool)))
    patient_target_list = list(range(len(solver.target)))
    np.random.seed(1234)
    for _ in range(1000):
        patients_pool = np.random.choice(
            patient_pool_list, replace=False, size=match_size
        )
        patients_target = np.random.choice(
            patient_target_list, replace=False, size=match_size
        )
        for f in range(n_features):
            pool_feature = sum_feature(solver.pool, patients_pool, f)
            target_feature = sum_feature(solver.target, patients_target, f)
            assert min_features[f] <= pool_feature <= max_features[f]
            assert min_features[f] <= target_feature <= max_features[f]
            abs_delta = abs(pool_feature - target_feature)
            assert abs_delta <= max_deltas[f]



import setuptools
from pybalance import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("environments/requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="pybalance",
    version=__version__,
    author="IEG Data Science",
    author_email="author@example.com",
    description="Population Matching",
    long_description=long_description,
    install_requires=requirements,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from ortools.sat.python import cp_model
import multiprocessing
import time
import numpy as np
import pandas as pd
from typing import List, Optional, Union
from sklearn.preprocessing import MinMaxScaler

import logging

logger = logging.getLogger(__name__)

from pybalance.propensity import PropensityScoreMatcher
from pybalance.utils import (
    MatchingData,
    BaseBalanceCalculator,
    BalanceCalculator,
    split_target_pool,
)


def _check_fitted(matcher):
    if matcher.best_match is None:
        raise (ValueError, "Matcher has not been fitted!")


def compute_truncation_error(x: np.ndarray) -> float:
    return np.abs(x.astype(int) - x).sum() / np.abs(x).sum()


def _rescale_for_discretization(
    target: np.array, pool: np.array, tolerance: float = 0.01
):
    """
    Find a scale factor that allows one to convert datasets to integer datatypes
    with minimal loss. Input parameter tolerance specifies the max allowable
    truncation error as a fraction.

    tolerance has a somewhat surprising effect of requiring more memory to
    search. by using a lower tolerance, one can put better bounds on the
    optimization variables and I think this helps memory usage.
    """

    # Get initial truncation error
    truncation_error = max(
        compute_truncation_error(target), compute_truncation_error(pool)
    )

    # Ideally, no scaling is required at all, e.g., if everything is already
    # an integer. Start with an initial scaling of 1 and increase as needed
    # to find the right scale.
    scalefac = 1

    # Keep increasing the scale factor until the (worst) truncation error is
    # less than the specified tolerance.
    while truncation_error > tolerance:
        truncation_error = max(
            compute_truncation_error(target * scalefac),
            compute_truncation_error(pool * scalefac),
        )
        scalefac *= 1.2

    # We add an extra factor of 2 for good measure. Better safe than sorry
    scalefac *= 2
    logger.info(
        f"Scaling features by factor {scalefac:.2f} in order to use integer solver with <= {100 * truncation_error:.4f}% loss."
    )

    target = (scalefac * target).astype(int).tolist()
    pool = (scalefac * pool).astype(int).tolist()

    return target, pool


class SolutionPrinter(cp_model.CpSolverSolutionCallback):
    """Print intermediate solutions."""

    def __init__(self, ind, deltas, matcher):
        cp_model.CpSolverSolutionCallback.__init__(self)
        self.__ind = ind
        self.__deltas = deltas
        self.__solution_count = 0
        self.__start_time = time.time()
        self.solutions = []
        self.balance_scores = []
        self.matcher = matcher
        self.balance_calculator = matcher.balance_calculator
        if self.balance_calculator is not None:
            logger.info(
                f"Initial balance score: {self.balance_calculator.distance(list(range(len(self.balance_calculator.pool)))):.4f}"
            )

    def on_solution_callback(self):
        current_time = time.time()
        logger.info("=========================================")
        logger.info(
            "Solution %i, time = %0.2f m"
            % (self.__solution_count + 1, (current_time - self.__start_time) / 60)
        )
        logger.info(f"Objective:\t{self.ObjectiveValue()}")

        indices = []
        for i in range(len(self.__ind)):
            if self.Value(self.__ind[i]):
                indices.append(i)
        self.solutions.append(indices)

        if self.balance_calculator is not None:
            # ---- Get indices of patients ----#

            balance = self.balance_calculator.distance(indices)
            self.balance_scores.append(balance)
            logger.info(f"Balance:\t{balance:.4f}")

        # ---- Print Deltas per feature ----#
        logger.info(f"Patients:\t{sum([self.Value(x) for x in self.__ind])}")

        # ---- Print Deltas per feature ----#
        # Don't touch this code. You may think it is written by a crazy person,
        # but I promise ORTools will crash if you try to change it. I have no
        # idea why.
        headers = self.balance_calculator.preprocessor.output_headers["all"]
        j = 0
        for i in self.__deltas:
            feat = headers[j]
            logger.debug(f"Feature deltas: %12i {feat} " % self.Value(i))
            j += 1
        logger.info(" ")

        self.__solution_count += 1

    def solution_count(self):
        return self.__solution_count


class ConstraintSatisfactionMatcher(object):
    """
    Constructs the class patient selection class.

    :param matching_data: A MatchingData object describing the pool and target
        populations. See utils.matching_data.

    :param objective: Matching objective to optimize. Technically, you can pass
        any balance calculator, but this solver cannot handle non-linear
        objective functions. The solver uses the preprocessing from the balance
        calculator for setting up the problem; the balance calculator itself is
        used to report the balance of generated matches but not in actually
        finding solutions (since the CS solver needs a discretized objective
        function). The solver will optimize the absolute mean difference on the
        output features of the balance calculator's preprocessing.

    :param match_size: How large the match population should be. If passed as a
        float, interpreted to be a fraction of the target population.

    :param time_limit: Time limit to stop solving in seconds (def: 180 sec).

    :param num_workers: Number of workers to use to optimize objective. See
        https://github.com/google/or-tools/blob/stable/ortools/sat/sat_parameters.proto#L556
        for more detail.

    :param ps_hinting: Compute a propensity score match and use the result as a
        hint for the solver

    :param verbose: Verbose solving.
    """

    def __init__(
        self,
        matching_data: MatchingData,
        objective: Union[str, BaseBalanceCalculator] = "beta",
        match_size: int = None,
        time_limit: float = 180,
        num_workers: int = 4,
        ps_hinting: bool = False,
        verbose: bool = True,
    ):
        self.matching_data = matching_data.copy()
        self.orig_target, self.orig_pool = split_target_pool(self.matching_data)

        if isinstance(objective, str):
            self.balance_calculator = BalanceCalculator(self.matching_data, objective)
            self.objective = objective
        else:
            self.balance_calculator = objective
            self.objective = self.balance_calculator.name

        target, pool = (
            self.balance_calculator.target.cpu().numpy(),
            self.balance_calculator.pool.cpu().numpy(),
        )
        # We will have to rescale everything to use an integer solver. If
        # features are not normalized, then features with larger overall scale
        # end up being emphasized more in the objective function.
        scaler = MinMaxScaler()
        target = scaler.fit_transform(target)
        pool = scaler.transform(pool)
        self.n_target, self.n_pool = len(target), len(pool)
        self.n_features = len(target[0])
        self.weights = self.get_weights()
        self.target, self.pool = _rescale_for_discretization(
            target * self.weights, pool * self.weights, tolerance=0.01
        )

        # ===== Sanity Check ======#
        assert len(self.target[0]) == len(
            self.pool[0]
        ), "The number of features should be the same in both reference population and input!"
        assert (
            self.n_target < self.n_pool
        ), "Number of patients in treatment arm should be equal or less then control."

        self.match_size = match_size
        self.time_limit = time_limit
        self.num_workers = num_workers
        self.ps_hinting = ps_hinting
        self.verbose = verbose
        self._check_params()

        self.ref_features = [
            sum([t[f] for t in self.target]) for f in range(self.n_features)
        ]
        self._reset_best_match()

    def _check_params(self):
        # set match size
        if self.match_size is not None:
            if not self.match_size <= self.n_target:
                raise ValueError(
                    "Match size must not be greater than target population size."
                )
            if self.match_size <= 1:
                # Assume it represents a fraction of the target size
                self.match_size = int(self.match_size * self.n_target)
        else:
            self.match_size = self.n_target

    def get_params(self):
        params = [
            "objective",
            "match_size",
            "time_limit",
            "num_workers",
            "ps_hinting",
            "verbose",
        ]
        return dict((p, getattr(self, p)) for p in params)

    def _reset_best_match(self):
        self.best_match = None
        self.best_match_idx = None
        self.best_score = np.inf

    def get_weights(self):
        if hasattr(self.balance_calculator.preprocessor, "feature_weights"):
            weights = self.balance_calculator.preprocessor.feature_weights.cpu().numpy()
            min_nonzero_weight = min([w for w in weights if w > 0])
            weights = [int(10 * w / min_nonzero_weight) for w in weights]
        else:
            weights = [1] * self.n_features

        headers = self.balance_calculator.preprocessor.output_headers["all"]
        for feat, weight in zip(headers, weights):
            logger.debug(f"Feature weights: {weight:6d} {feat}")

        return weights

    def _get_limits(self, f, match_size=None):
        if match_size is None:
            match_size = self.n_target

        weight = self.weights[f]

        # sort feature values ascending
        pool_features = sorted([p[f] for p in self.pool])
        min_pool_feature = sum(pool_features[:match_size])
        max_pool_feature = sum(pool_features[-match_size:])

        target_feature_values = sorted([p[f] for p in self.target])
        min_target_feature = sum(target_feature_values[:match_size])
        max_target_feature = sum(target_feature_values[-match_size:])

        max_delta = weight * max(
            abs(max_pool_feature - min_target_feature),
            abs(max_target_feature - min_pool_feature),
        )

        if max_delta >= 2**50:
            logger.warning(f"Inferred max_delta = {max_delta}.")
            logger.warning(
                "Feature dynamic range may exceed integer dynamic range and lead to suboptimal solutions."
            )

        min_feature = min(min_target_feature, min_pool_feature)
        max_feature = max(max_target_feature, max_pool_feature)

        feat = self.balance_calculator.preprocessor.output_headers["all"][f]
        logger.debug(f"Feature limits:  [{min_feature}, {max_feature}] {feat}")

        return min_feature, max_feature, max_delta

    def match(self, hint: Optional[List[int]] = None) -> MatchingData:
        """
        Match populations passed during __init__(). Returns MatchingData
        instance containing the matched pool and target populations.

        :param hint: You can supply a "hint" as either (1) A list of indices to
            the pool. It will be assumed that the entire target is used, or (2)
            A list of two lists, the first list being the indices to the target,
            and the second being the indices to the pool, or (3) by omitting the
            hint altoghether and passing ps_hinting=True in __init__(). In case
            (3), a propensity score model will be estimated on the fly and used
            to create a match population as a hint to the solver.

            I admit the interface here is a bit confusing. We will clean this
            up in a later release.
        """
        time_limit = self.time_limit

        logger.info(
            f"Solving for match population with size = {self.match_size}. Will match on {self.n_features} dimensions ..."
        )

        # ========= Create Model===========#
        model = cp_model.CpModel()

        # ========= Variables===========#
        # --- Binary flag for inclusion of patients ---------#
        x = []  # pool
        y = []  # target
        for i in range(self.n_pool):
            x.append(model.NewBoolVar(f"x[{i}]"))
        for i in range(self.n_target):
            y.append(model.NewBoolVar(f"y[{i}]"))

        logger.info("Calculating bounds on feature variables ...")

        abs_deltas = []
        features = []
        ref_features = []
        deltas = []
        for i in range(self.n_features):
            min_feature, max_feature, max_delta = self._get_limits(i, self.match_size)
            abs_deltas.append(model.NewIntVar(0, max_delta, f"abs_deltas[{i}]"))
            features.append(model.NewIntVar(min_feature, max_feature, f"features[{i}]"))
            deltas.append(model.NewIntVar(-max_delta, max_delta, f"deltas[{i}]"))
            ref_features.append(
                model.NewIntVar(min_feature, max_feature, f"ref_features[{i}]")
            )

        logger.info("Applying constraints ...")

        # ========= Constraints===========#
        for j in range(self.n_features):
            # --- aggregate for feature j ---------#
            model.Add(
                sum(self.pool[i][j] * x[i] for i in range(self.n_pool)) == features[j]
            )
            model.Add(
                sum(self.target[i][j] * y[i] for i in range(self.n_target))
                == ref_features[j]
            )
            # --- Taxicab distance ---------#
            model.Add(self.weights[j] * (features[j] - ref_features[j]) == deltas[j])
            model.AddAbsEquality(abs_deltas[j], deltas[j])

        # --- Cardinality ---------#
        model.Add(sum(x[i] for i in range(self.n_pool)) == self.match_size)
        model.Add(sum(y[i] for i in range(self.n_target)) == self.match_size)

        # ========= Hint! ==============#
        if self.ps_hinting or hint is not None:
            matching_data = self.matching_data.copy()
            target, pool = split_target_pool(matching_data)
            target_name = target[matching_data.population_col].unique()[0]
            pool_name = pool[matching_data.population_col].unique()[0]

            if hint is None:
                logger.info("Training PS model as guide for solver ...")

                target.loc[:, "ix"] = list(range(len(target)))
                pool.loc[:, "ix"] = list(range(len(pool)))
                matching_data = MatchingData(
                    pd.concat([target, pool]), headers=self.matching_data.headers
                )
                ps_matcher = PropensityScoreMatcher(
                    matching_data, objective=self.objective
                )
                ps_match = ps_matcher.match()

                target_hint = (
                    ps_match.data[ps_match.data.population == target_name]
                    .ix.astype(int)
                    .tolist()
                )
                pool_hint = (
                    ps_match.data[ps_match.data.population == pool_name]
                    .ix.astype(int)
                    .tolist()
                )
            else:
                if len(hint) == 2:
                    target_hint, pool_hint = hint
                else:
                    target_hint = [1] * len(target)
                    pool_hint = hint

            # Kind of sanity check that the indices are aligned between what
            # comes out of the PS solver and what we will apply as hints
            obj = 0
            for j in range(self.n_features):
                feat = sum(self.pool[i][j] for i in pool_hint)
                ref = sum(self.target[i][j] for i in target_hint)
                delta = self.weights[j] * abs(feat - ref)
                obj += delta
            logger.info(f"Hint achieves objective value = {obj}.")

            logger.info("Applying hints ...")
            for i, var in enumerate(x):
                if i in pool_hint:
                    model.AddHint(var, 1)
                else:
                    model.AddHint(var, 0)

            for i, var in enumerate(y):
                if i in target_hint:
                    model.AddHint(var, 1)
                else:
                    model.AddHint(var, 0)

        # ========= Objective===========#
        model.Minimize(sum(abs_deltas))

        # ===== Creates a solver =======#
        solver = cp_model.CpSolver()
        if time_limit is not None:
            solver.parameters.max_time_in_seconds = time_limit

        solver.parameters.num_workers = min(
            self.num_workers, multiprocessing.cpu_count()
        )
        if solver.parameters.num_workers != 1:
            solver.parameters.log_search_progress = False
            solver.parameters.share_objective_bounds = True
            solver.parameters.share_level_zero_bounds = True

        logger.info("Solving with %d workers ..." % solver.parameters.num_workers)
        # ========= Solve===========#
        # ----- Append the important stuff for printing -----#

        if self.verbose:
            solution_printer = SolutionPrinter(x, abs_deltas, self)
            status = solver.SolveWithSolutionCallback(model, solution_printer)
            logger.info("Status = %s" % solver.StatusName(status))
            logger.info(
                "Number of solutions found: %i" % solution_printer.solution_count()
            )
        else:
            status = solver.Solve(model)

        if solver.StatusName(status) in ["FEASIBLE", "OPTIMAL"]:
            pool_indices = [j for j, i in enumerate(x) if solver.Value(i)]
            pool = self.orig_pool.iloc[pool_indices]
            target_indices = [j for j, i in enumerate(y) if solver.Value(i)]
            target = self.orig_target.iloc[target_indices]
            match = MatchingData(
                data=pd.concat([target, pool]),
                headers=self.matching_data.headers,
                population_col=self.matching_data.population_col,
            )
        else:
            match = self.matching_data

        self.best_match = match

        return match

    def get_best_match(self):
        _check_fitted(self)
        return self.best_match



import numpy as np

from pybalance.genetic import GeneticMatcher
from pybalance.sim import generate_toy_dataset


def test_noduplicates():
    # test that no patient is duplicated in candidate populations generated by generate_offspring()

    m = generate_toy_dataset()
    matcher = GeneticMatcher(m, objective="beta")
    matcher._init_first_generation()
    matcher._check_stopping_conditions()
    for generation in range(3):
        matcher._generate_offspring()
        for pop in matcher.candidate_populations:
            assert len(pop) == len(np.unique(pop))



from pybalance.sim import rng


def test_truncated_norm_rct():
    data = rng.generate_random_feature_data_rct(size=1000)

    assert data.age.min() >= 18
    assert data.age.max() <= 75
    assert data.height.min() >= 125
    assert data.height.max() <= 195


def test_truncated_norm_rwd():
    data = rng.generate_random_feature_data_rwd(size=1000)

    assert data.age.min() >= 18
    assert data.age.max() <= 75
    assert data.height.min() >= 125
    assert data.height.max() <= 195



# generate simulated data
import os
import numpy as np
from scipy.stats import truncnorm, uniform
import pandas as pd
from pybalance.utils import MatchingData


def generate_truncated_distributions(rng, mn, mx, size):
    """
    Generate multivariate normal variables with truncations.
        rng = a function which takes as only parameter size and returns
            random variations of shape size x N
        cov = covariance matrix, shape N x N
        mn = lower limits, shape N x 1
        mx = upper limits, shape N x 1
        size = number of samples to draw
    This is implemented by rejection sampling, so if the code runs
    for a really long time, you have probably set parameters for which
    most samples get rejected.
    """
    out = np.empty((0, len(mn)))
    while out.shape[0] < size:
        _out = rng(2 * size)
        where = ((_out <= mx) & (_out >= mn)).min(axis=1)
        if len(where):
            _out = _out[where]
            out = np.vstack([out, _out])

    return out[:size, :]


def multivariate_truncnorm(mu, cov, mn, mx, size):
    """
    Generate multivariate normal variables with truncations.
        mu = means of the covariates, shape N x 1
        cov = covariance matrix, shape N x N
        mn = lower limits, shape N x 1
        mx = upper limits, shape N x 1
        size = number of samples to draw
    This is implemented by rejection sampling, so if the code runs
    for a really long time, you have probably set parameters for which
    most samples get rejected.
    """

    def rng(x):
        return np.random.multivariate_normal(mu, cov, size=x)

    return generate_truncated_distributions(rng, mn, mx, size)


def _truncnorm(mn, mx, mu, std, size):
    """
    For reasons I don't understand, the loc and scale parameters in truncnorm
    don't behave as I expect. Here I reimplment the interface to truncnrom to
    make more sense.
    """
    a, b = (mn - mu) / std, (mx - mu) / std
    return mu + std * truncnorm.rvs(a, b, size=size)


def correlate_vars(x, y, cor):
    """
    Given two uncorrelated covariates x and y return a new random variable with the same
    mean and standard deviation as y but with correlation cor to covariate x.
    """
    mu_0 = y.mean()
    sigma_0 = y.std()
    y = cor * x / x.std() + np.sqrt(1 - cor**2) * y / y.std()
    y = (sigma_0 / y.std()) * (y - y.mean()) + mu_0
    return y


def generate_random_feature_data_rct(
    size, mu_h=150, std_h=20, cor_hw=-0.8, cor_ag=0.8, p_binary=[], seed=42
):
    np.random.seed(seed)

    def rng(size):
        weight = uniform.rvs(loc=50, scale=120 - 50, size=size)
        height = np.random.normal(loc=mu_h, scale=std_h, size=size)
        height = correlate_vars(weight, height, cor_hw)
        return np.vstack([weight, height]).T

    wh = generate_truncated_distributions(rng, mn=[50, 125], mx=[120, 195], size=size)
    weight = wh[:, 0].T
    height = wh[:, 1].T

    def rng(size):
        gender = np.random.choice([0, 1], size=size, replace=True, p=[0.5, 0.5])
        age = np.random.normal(loc=50, scale=20, size=size)
        age = correlate_vars(gender, age, cor_ag)
        return np.vstack([gender, age]).T

    ga = generate_truncated_distributions(rng, mn=[0, 18], mx=[1, 75], size=size)
    gender = ga[:, 0].T
    age = ga[:, 1].T

    haircolor = np.random.choice([0, 1, 2], size=size, replace=True, p=[0.2, 0.4, 0.4])
    country = np.random.choice(
        [0, 1, 2, 3, 4, 5], size=size, replace=True, p=[0.1, 0.2, 0.2, 0.1, 0.2, 0.2]
    )

    data = pd.DataFrame.from_dict(
        {
            "age": age,
            "height": height,
            "weight": weight,
            "gender": gender,
            "haircolor": haircolor,
            "country": country,
            "population": ["target"] * size,
        }
    )
    for i, p in enumerate(p_binary):
        data.loc[:, f"binary_{i}"] = np.array(np.random.rand(size) < p).astype(int)

    return data


def generate_random_feature_data_rwd(
    size, mu_w=90, std_w=20, cor_hw=0.8, cor_ag=-0.8, p_binary=[], seed=43
):
    np.random.seed(seed)

    def rng(size):
        weight = np.random.normal(loc=mu_w, scale=std_w, size=size)
        height = uniform.rvs(loc=125, scale=195 - 125, size=size)
        weight = correlate_vars(height, weight, cor_hw)
        return np.vstack([weight, height]).T

    wh = generate_truncated_distributions(rng, mn=[50, 125], mx=[120, 195], size=size)
    weight = wh[:, 0].T
    height = wh[:, 1].T

    def rng(size):
        gender = np.random.choice([0, 1], size=size, replace=True, p=[0.6, 0.4])
        age = np.random.normal(loc=65, scale=20, size=size)
        age = correlate_vars(gender, age, cor_ag)
        return np.vstack([gender, age]).T

    ga = generate_truncated_distributions(rng, mn=[0, 18], mx=[1, 75], size=size)
    gender = ga[:, 0].T
    age = ga[:, 1].T

    haircolor = np.random.choice([0, 1, 2], size=size, replace=True, p=[0.4, 0.3, 0.3])
    country = np.random.choice(
        [0, 1, 2, 3, 4, 5], size=size, replace=True, p=[0, 0.1, 0.2, 0.3, 0.3, 0.1]
    )

    data = pd.DataFrame.from_dict(
        {
            "age": age,
            "height": height,
            "weight": weight,
            "gender": gender,
            "haircolor": haircolor,
            "country": country,
            "population": ["pool"] * size,
        }
    )

    for i, p in enumerate(p_binary):
        data.loc[:, f"binary_{i}"] = np.array(np.random.rand(size) < p).astype(int)

    return data


def generate_toy_dataset(n_pool=10000, n_target=1000, seed=45):
    """
    Generate a toy matching dataset with n_pool patients in the pool and
    n_target patients in the target population. For finer control, see
    generate_random_feature_data_rwd and generate_random_feature_data_rct.
    """
    pool = generate_random_feature_data_rwd(
        n_pool, cor_hw=0.5, cor_ag=-0.5, p_binary=[0.1, 0.3, 0.5, 0.8], seed=seed
    )
    target = generate_random_feature_data_rct(
        n_target, std_h=20, cor_ag=0.5, p_binary=[0.3, 0.5, 0.3, 0.5], seed=seed + 1
    )
    feature_data = pd.concat([pool, target])
    feature_data.loc[:, "patient_id"] = list(range(len(feature_data)))
    return MatchingData(feature_data)


def load_paper_dataset():
    """
    Load the simulated matching dataset presented in the pybalance paper.
    """
    filepath = "pool250000-target25000-normal0-lognormal0-binary4.parquet"
    resource = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data", filepath
    )
    return MatchingData(resource)
